# Form implementation generated from reading ui file 'registersubmoduledialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_RegisterSubmoduleDialog(object):
    def setupUi(self, RegisterSubmoduleDialog):
        RegisterSubmoduleDialog.setObjectName("RegisterSubmoduleDialog")
        RegisterSubmoduleDialog.setWindowModality(Qt.WindowModality.WindowModal)
        RegisterSubmoduleDialog.resize(444, 221)
        self.formLayout = QFormLayout(RegisterSubmoduleDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.pathLabel = QLabel(parent=RegisterSubmoduleDialog)
        self.pathLabel.setObjectName("pathLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.pathLabel)
        self.pathValue = QElidedLabel(parent=RegisterSubmoduleDialog)
        self.pathValue.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pathValue.sizePolicy().hasHeightForWidth())
        self.pathValue.setSizePolicy(sizePolicy)
        self.pathValue.setText("PATH")
        self.pathValue.setObjectName("pathValue")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.pathValue)
        self.nameLabel = QLabel(parent=RegisterSubmoduleDialog)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.remoteLabel = QLabel(parent=RegisterSubmoduleDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteLabel.sizePolicy().hasHeightForWidth())
        self.remoteLabel.setSizePolicy(sizePolicy)
        self.remoteLabel.setObjectName("remoteLabel")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.remoteLabel)
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.remoteComboBox = QComboBoxWithPreview(parent=RegisterSubmoduleDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteComboBox.sizePolicy().hasHeightForWidth())
        self.remoteComboBox.setSizePolicy(sizePolicy)
        self.remoteComboBox.setObjectName("remoteComboBox")
        self.horizontalLayout_3.addWidget(self.remoteComboBox)
        self.remoteHint = QHintButton(parent=RegisterSubmoduleDialog)
        self.remoteHint.setObjectName("remoteHint")
        self.horizontalLayout_3.addWidget(self.remoteHint)
        self.formLayout.setLayout(2, QFormLayout.ItemRole.FieldRole, self.horizontalLayout_3)
        spacerItem = QSpacerItem(20, 1, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding)
        self.formLayout.setItem(3, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.absorbExplainer = QLabel(parent=RegisterSubmoduleDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.absorbExplainer.sizePolicy().hasHeightForWidth())
        self.absorbExplainer.setSizePolicy(sizePolicy)
        self.absorbExplainer.setWordWrap(True)
        self.absorbExplainer.setObjectName("absorbExplainer")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.SpanningRole, self.absorbExplainer)
        self.buttonBox = QDialogButtonBox(parent=RegisterSubmoduleDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(5, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.nameEdit = QLineEdit(parent=RegisterSubmoduleDialog)
        self.nameEdit.setObjectName("nameEdit")
        self.horizontalLayout_5.addWidget(self.nameEdit)
        self.nameHint = QHintButton(parent=RegisterSubmoduleDialog)
        self.nameHint.setObjectName("nameHint")
        self.horizontalLayout_5.addWidget(self.nameHint)
        self.formLayout.setLayout(1, QFormLayout.ItemRole.FieldRole, self.horizontalLayout_5)
        self.nameLabel.setBuddy(self.nameEdit)
        self.remoteLabel.setBuddy(self.remoteComboBox)

        self.retranslateUi(RegisterSubmoduleDialog)
        self.buttonBox.accepted.connect(RegisterSubmoduleDialog.accept)
        self.buttonBox.rejected.connect(RegisterSubmoduleDialog.reject)
        RegisterSubmoduleDialog.setTabOrder(self.nameEdit, self.remoteComboBox)
        RegisterSubmoduleDialog.setTabOrder(self.remoteComboBox, self.nameHint)
        RegisterSubmoduleDialog.setTabOrder(self.nameHint, self.remoteHint)

    def retranslateUi(self, RegisterSubmoduleDialog):
        RegisterSubmoduleDialog.setWindowTitle(_p("RegisterSubmoduleDialog", "Register submodule"))
        self.pathLabel.setText(_p("RegisterSubmoduleDialog", "Workdir:"))
        self.nameLabel.setText(_p("RegisterSubmoduleDialog", "&Name:"))
        self.remoteLabel.setText(_p("RegisterSubmoduleDialog", "&Remote:"))
        self.remoteHint.setToolTip(_p("RegisterSubmoduleDialog", "<p>This will be the default remote for this submodule. People who clone your repo will use this remote to download the submodule."))
        self.absorbExplainer.setText(_p("RegisterSubmoduleDialog", "Once absorbed as a submodule, this copy of {sub} cannot be used independently from its superproject, {super}. This cannot be undone!"))
        self.nameHint.setToolTip(_p("RegisterSubmoduleDialog", "<p>This name uniquely identifies your submodule in “.gitmodules”. It can be anything you like, but common practice is to simply name it after the submodule’s workdir."))
from gitfourchette.toolbox.qcomboboxwithpreview import QComboBoxWithPreview
from gitfourchette.toolbox.qelidedlabel import QElidedLabel
from gitfourchette.toolbox.qhintbutton import QHintButton



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox.pathutils import compactPath
from gitfourchette.toolbox.qelidedlabel import QElidedLabel
from gitfourchette.toolbox.qhintbutton import QHintButton
from gitfourchette.toolbox.qtutils import tweakWidgetFont, DisableWidgetUpdatesContext
from gitfourchette.toolbox.textutils import escamp


class QFilePickerCheckBox(QWidget):
    def __init__(self, parent, caption: str = ""):
        super().__init__(parent)

        self.cachedPath = ""

        self.checkBox = QCheckBox(caption)
        self.pathLabel = QElidedLabel(self)
        self.pathLabel.setTextFormat(Qt.TextFormat.PlainText)
        self.browseButton = QToolButton()
        self.browseButton.setText(_p("button to open a file browser to select a file", "Select…"))

        self.warningButton = QHintButton(self, iconKey="achtung")

        layout = QGridLayout(self)
        layout.setContentsMargins(QMargins())
        layout.setVerticalSpacing(0)
        layout.addWidget(self.checkBox,         0, 0, 1, 3)
        layout.addWidget(self.browseButton,     1, 0)
        layout.addWidget(self.warningButton,    1, 1)
        layout.addWidget(self.pathLabel,        1, 2)
        layout.setColumnStretch(2, 1)

        self.checkBox.toggled.connect(self.autoBrowse)
        self.browseButton.clicked.connect(self.browse)
        tweakWidgetFont(self.pathLabel, 90)
        f = tweakWidgetFont(self.browseButton, 90)
        lh = QFontMetrics(f).height() + 2
        self.browseButton.setMaximumHeight(lh)
        self.warningButton.setMaximumHeight(lh)
        self.updateControls()

    def fileDialog(self) -> QFileDialog:
        """ Override this function to bring up a custom QFileDialog. """
        raise NotImplementedError("override this")

    def validatePath(self, path: str) -> str:
        """ Override this function to display a validation warning when the user selects a file.
        Returning an empty string causes the warning to disappear. """
        raise NotImplementedError("override this")

    def setText(self, caption: str):
        self.checkBox.setText(caption)

    def isChecked(self) -> bool:
        return self.checkBox.isChecked()

    def path(self) -> str:
        return self.cachedPath if self.isChecked() else ""

    def autoBrowse(self):
        """ If checkbox is ticked and path is empty, bring up file browser. """
        if self.isChecked() and not self.cachedPath:
            self.browse()
        else:
            self.updateControls()

    def browse(self):
        qfd = self.fileDialog()
        qfd.fileSelected.connect(self.setPath)
        qfd.rejected.connect(self.onRejectFileDialog)
        qfd.show()

    def setPath(self, path: str):
        self.cachedPath = path
        self.warningButton.setToolTip(self.validatePath(path))

        with QSignalBlocker(self.checkBox):
            self.checkBox.setChecked(bool(path))
        self.updateControls()

    def onRejectFileDialog(self):
        # File browser canceled and lineedit empty, untick checkbox
        if not self.cachedPath:
            self.checkBox.setChecked(False)

    @DisableWidgetUpdatesContext.methodDecorator
    def updateControls(self):
        path = self.path()
        hasPath = bool(path)
        self.pathLabel.setText(escamp(compactPath(path)))
        self.browseButton.setVisible(hasPath)
        self.pathLabel.setVisible(hasPath)
        self.warningButton.setVisible(hasPath and bool(self.warningButton.toolTip()))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import bisect
import logging
from dataclasses import dataclass
from collections.abc import Iterable, Iterator, Set
from typing import ClassVar, Protocol

from gitfourchette.appconsts import *
from gitfourchette.porcelain import Oid, Signature

logger = logging.getLogger(__name__)


class CommitTraits(Protocol):
    id: Oid
    parent_ids: list[Oid]
    message: str
    author: Signature
    committer: Signature
    gpg_signature: tuple[bytes, bytes]


KF_INTERVAL = 5000
"""
Interval (in number of commits) at which keyframes are saved while preparing
the graph.

The bigger the interval...:
- faster initial loading of the repo & less memory usage;
- but slower random access to any point of the graph.
"""

DEAD_VALUE = Oid(hex="deaddeaddeaddeaddeaddeaddeaddeaddeaddead")


@dataclass(frozen=True)
class BatchRow:
    """
    For a row in the Graph, BatchRow keeps track of the row's batch number
    and position within the batch.

    When a repo is refreshed, new commits may appear at the top of the graph.
    This may push existing rows down, away from the top of the graph.

    Recomputing row numbers for the entire graph becomes costly if it contains
    tens of thousands of commits.

    Instead, we keep track of batches of rows inserted into the graph at each
    refresh. We adjust a single offset value for each batch.
    """

    b: int = -1
    "Batch number (look up its global offset in BatchManager)"

    y: int = -1
    "Position of this row relative to its batch"

    class BatchManager:
        """
        Manages row batch offsets.

        A single BatchManager is shared by all repositories opened throughout
        the lifespan of the app.
        """

        globalOffsets: ClassVar[list[int]] = []
        freeBatchNos: ClassVar[list[int]] = []

        @classmethod
        def reserveNewBatch(cls):
            if cls.freeBatchNos:
                batchNo = cls.freeBatchNos.pop()
                cls.globalOffsets[batchNo] = 0
            else:
                cls.globalOffsets.append(0)
                batchNo = len(cls.globalOffsets) - 1

            return batchNo

        @classmethod
        def freeBatch(cls, batchNo: int):
            cls.globalOffsets[batchNo] = -0xDEADBEEF
            cls.freeBatchNos.append(batchNo)

            # Compact freed batches at end of list
            for bn in range(len(cls.globalOffsets) - 1, -1, -1):
                assert bn == len(cls.globalOffsets) - 1
                if bn in cls.freeBatchNos:
                    cls.freeBatchNos.remove(bn)
                    cls.globalOffsets.pop()
                else:
                    # Stop iterating once the last batch is non-free
                    break

        @classmethod
        def shiftBatches(cls, shift: int, batchNos: Iterable[int]):
            for b in batchNos:
                assert b not in cls.freeBatchNos, "trying to shift a freed batch"
                cls.globalOffsets[b] += shift

    def isValid(self):
        """
        A BatchRow is considered valid if it is part of a valid batch, it has a
        positive offset in the batch, and it resolves to a positive int.
        """
        return self.b >= 0 and self.y >= 0 and int(self) >= 0

    def __repr__(self):
        offset = -1 if self.b < 0 else BatchRow.BatchManager.globalOffsets[self.b]
        return f"{int(self)} (b{self.b}r{self.y}+{offset})"

    def __str__(self):
        return str(int(self))

    def __int__(self) -> int:
        """Any BatchRow is convertible to int, giving a global row index.
        Note that the int value of any given BatchRow may change over time
        as the graph gets spliced."""
        if self.b < 0:
            return -1
        return BatchRow.BatchManager.globalOffsets[self.b] + self.y

    # -------------------------------------------------------------------------
    # Arithmetics

    def __add__(self, other: BatchRow | int) -> int:
        return int(self) + int(other)

    def __sub__(self, other: BatchRow | int) -> int:
        return int(self) - int(other)

    def __mod__(self, other: BatchRow | int) -> int:
        return int(self) % int(other)

    def __rmod__(self, other: BatchRow | int) -> int:
        return int(other) % int(self)

    # -------------------------------------------------------------------------
    # Comparisons
    # (BatchRows must be comparable so we can use bisect.)

    def __le__(self, other: BatchRow | int) -> bool:
        return int(self) <= int(other)

    def __lt__(self, other: BatchRow | int) -> bool:
        return int(self) < int(other)

    def __ge__(self, other: BatchRow | int) -> bool:
        return int(self) >= int(other)

    def __gt__(self, other: BatchRow | int) -> bool:
        return int(self) > int(other)

    def __eq__(self, other: object) -> bool:
        try:
            return int(self) == int(other)  # type: ignore[call-overload]
        except (TypeError, ValueError):
            return False


BATCHROW_UNDEF = BatchRow(-1, -1)
"Use this special BatchRow as a placeholder for a row position that is yet to be determined."


@dataclass
class ChainHandle:
    """ Object shared by arcs on the same chain. """
    _t: BatchRow = BATCHROW_UNDEF
    _b: BatchRow = BATCHROW_UNDEF
    alias: ChainHandle | None = None

    def isValid(self):
        # It's OK for bottomRow to be dangling, but not topRow.
        return self.topRow.isValid()

    def __repr__(self):
        tr = int(self.topRow)
        br = int(self.bottomRow)
        return f"Chain({tr}\u2192{br})"

    def resolve(self):
        """ Return non-aliased ChainHandle """

        if self.alias is None:
            return self

        if self.alias.alias is not None:
            # Flatten aliases
            frontier = [self, self.alias]
            root = frontier[-1]
            while True:
                root = root.alias
                if root.alias is None:
                    break
                frontier.append(root)
            for ch in frontier:
                ch.alias = root
            if APP_DEBUG:
                assert root.alias is None
                assert root not in frontier
                assert root is self.alias

        return self.alias

    @property
    def topRow(self) -> BatchRow:
        return self.resolve()._t

    @property
    def bottomRow(self) -> BatchRow:
        return self.resolve()._b

    @bottomRow.setter
    def bottomRow(self, value):
        self._b = value
        assert self.alias is None, "Cannot modify row in aliased ChainHandle"

    def setAliasOf(self, master: ChainHandle):
        if self.alias is not None:
            self.alias.setAliasOf(master)

        self.alias = master
        self._t = BATCHROW_UNDEF
        self._b = BATCHROW_UNDEF


@dataclass
class ArcJunction:
    """ Represents the merging of an Arc into another Arc. """

    joinedAt: BatchRow
    "Row number in which this junction occurs"

    joinedBy: Oid
    "Hash of the joining arc's opening commit"

    def __lt__(self, other: ArcJunction):
        """
        Calling sorted() on a list of ArcJunctions will sort them by `joinedAt` (row numbers).
        """
        return self.joinedAt < other.joinedAt


@dataclass
class Arc:
    """ An arc connects two commits in the graph.

    Commits appear before their parents in the commit sequence.
    When processing the commit sequence, we "open" a new arc when
    encountering a new commit. The arc stays "open" until we've found
    the commit's parent in the sequence, at which point we "close" it.

    Other arcs may merge into an open arc via an ArcJunction.
    """

    openedAt: BatchRow
    "Row number in which this arc was opened"

    closedAt: BatchRow
    "Row number in which this arc was closed (may be BATCHROW_UNDEF until resolved)"

    chain: ChainHandle
    "Row number of the tip of the arc chain (topmost commit in branch)"

    lane: int
    "Lane assigned to this arc"

    openedBy: Oid | None
    "Hash of the opening commit (in git parlance, the child commit)"

    closedBy: Oid | None
    "Hash of the closing commit (in git parlance, the parent commit)"

    junctions: list[ArcJunction]
    "Other arcs merging into this arc"

    nextArc: Arc | None = None
    "Next node in the arc linked list"

    def __repr__(self):
        oa = int(self.openedAt)
        ca = int(self.closedAt)
        ob = str(self.openedBy)[:5]
        cb = str(self.closedBy)[:5]
        dangling = "?" if not self.closedAt.isValid() else ""
        return f"Arc({self.chain} {ob}\u2192{cb}{dangling} {oa}\u2192{ca})"

    def length(self):
        assert type(self.openedAt) is BatchRow
        assert type(self.closedAt) is BatchRow
        return int(self.closedAt) - int(self.openedAt)

    def __next__(self):
        return self.nextArc

    def __iter__(self) -> Iterator[Arc]:
        arc: Arc | None = self
        while arc is not None:
            yield arc
            assert arc.nextArc != arc, "self-referencing arc!"
            arc = arc.nextArc

    def getNumberOfArcsFromHere(self):
        n = 0
        for _ in self:
            n += 1
        return n

    def isParentlessCommitBogusArc(self):
        return self.openedBy == self.closedBy

    def isIndependentOfRowsAbove(self, row: int):
        """ Return True if this arc is entirely independent of row numbers lower than `row`. """
        return self.openedAt >= row and self.closedAt >= row

    def isStale(self, row: int):
        """
        Return True if this arc is considered stale at a given row in the graph.
        An arc is stale at row 'R' if it is closed above 'R',
        or if it is a ParentlessCommitBogusArc appearing at or above 'R',
        or if it is dangling (its closedAt row hasn't been resolved yet).
        """
        ca = int(self.closedAt)
        return (0 <= ca < row) or (0 <= ca == self.openedAt <= row)

    def isVisible(self, hiddenCommits: Set[Oid], row: int, filterJunctionRows=BatchRow.__lt__) -> bool:
        # FAIL if closing commit is hidden.
        if self.closedBy in hiddenCommits:
            return False

        # PASS if both the opening & closing commits are visible.
        if self.openedBy not in hiddenCommits:
            return True

        # Opening Commit hidden (above) and Closing Commit visible (below).
        # PASS if any of the junctions at/above the current row are visible.
        # TODO: We're looking at ALL the junctions - is it worth optimizing this?
        for j in self.junctions:
            if j.joinedBy in hiddenCommits:
                continue
            if filterJunctionRows(j.joinedAt, row):
                return True

        # All junctions should be hidden.
        return False


@dataclass
class Frame:
    """ A frame is a slice of the graph at a given row. """

    row: BatchRow
    commit: Oid | None
    solvedArcs: list[Arc | None]  # Arcs that have resolved their parent commit
    openArcs: list[Arc | None]  # Arcs that have not resolved their parent commit yet
    lastArc: Arc

    def arcsClosedByCommit(self, hiddenCommits: Set[Oid] | None = None):
        if APP_DEBUG:
            # Assume that all the arcs in solvedArcs are either None, or are closed by this commit.
            assert all(arc is None or arc.closedAt == self.row for arc in self.solvedArcs)
            assert all(arc is None or arc.closedBy == self.commit for arc in self.solvedArcs)

        gen = (arc for arc in self.solvedArcs if arc)

        if hiddenCommits:
            assert self.commit not in hiddenCommits, "calling this func is pointless if commit is hidden"
            row = int(self.row)
            gen = (arc for arc in gen if arc.isVisible(hiddenCommits, row))

        return gen

    def arcsOpenedByCommit(self, hiddenCommits: Set[Oid] | None = None):
        row = int(self.row)
        gen = (arc for arc in self.openArcs if arc and arc.openedAt == row)

        if hiddenCommits:
            assert self.commit not in hiddenCommits, "calling this func is pointless if commit is hidden"
            # Not looking at junctions here because the parents of a visible commit
            # are all supposed to be visible too.
            gen = (arc for arc in gen if arc.closedBy not in hiddenCommits)

        return gen

    def arcsPassingByCommit(self, hiddenCommits: Set[Oid] | None = None, filterJunctionRows=BatchRow.__lt__):
        row = int(self.row)
        gen = (arc for arc in self.openArcs if arc and arc.openedAt != row)

        if hiddenCommits:
            assert self.commit not in hiddenCommits, "calling this func is pointless if commit is hidden"
            gen = (arc for arc in gen if arc.isVisible(hiddenCommits, row, filterJunctionRows))

        return gen

    def junctionsAtCommit(self, hiddenCommits: Set[Oid]):
        row = int(self.row)
        for arc in self.arcsPassingByCommit(hiddenCommits, BatchRow.__eq__):
            # TODO: We're looking at all the junctions here, but isVisible (via getArcsPassingByCommit)
            #       just looked at the specific junction we were looking for.
            for j in arc.junctions:
                if j.joinedAt == row and j.joinedBy not in hiddenCommits:
                    assert j.joinedBy == self.commit
                    yield arc, j
                    break  # stop iterating on junctions, look at next arc

    def homeArc(self) -> Arc:
        leftmostClosed = next(self.arcsClosedByCommit(), None)
        leftmostOpened = next(self.arcsOpenedByCommit(), None)
        if leftmostOpened and leftmostClosed:
            return max(leftmostClosed, leftmostOpened, key=lambda a: a.lane)
        elif leftmostClosed:
            return leftmostClosed
        elif leftmostOpened:
            return leftmostOpened
        else:
            # It's a parentless + childless commit.
            # Implementation detail: for parentless commits, we create a bogus arc
            # as the "last" arc in the frame; we can get the lane from this bogus arc.
            assert self.lastArc.isParentlessCommitBogusArc()
            assert self.lastArc.openedBy == self.commit
            return self.lastArc

    def homeLane(self) -> int:
        return self.homeArc().lane

    def homeChain(self) -> ChainHandle:
        return self.homeArc().chain

    def sealCopy(self) -> Frame:
        """
        Generated frames must be "sealed off" to make them suitable
        for rendering or for use as keyframes.
        """
        solvedArcsCopy = self.solvedArcs.copy()
        openArcsCopy = self.openArcs.copy()

        # Move arcs that just got closed to solved list
        for lane, arc in enumerate(openArcsCopy):
            if arc and 0 <= arc.closedAt <= self.row:
                openArcsCopy[lane] = None

                # For parentless commits, prevent bogus auto-closing arcs
                # from overwriting legitimate closed arcs.
                if not arc.isParentlessCommitBogusArc():
                    self.reserveArcListCapacity(solvedArcsCopy, lane+1)
                    solvedArcsCopy[lane] = arc

        # Remove stale closed arcs and trim "Nones" off end of list
        self.cleanUpArcList(openArcsCopy, self.row, alsoTrimBack=True)
        self.cleanUpArcList(solvedArcsCopy, self.row, alsoTrimBack=True)

        # In debug mode, make sure none of the arcs are dangling
        if APP_DEBUG:
            assert all(arc is None or arc.openedBy != DEAD_VALUE for arc in openArcsCopy)
            assert all(arc is None or arc.openedBy != DEAD_VALUE for arc in solvedArcsCopy)
            assert all(arc is None or arc.chain.isValid() for arc in openArcsCopy)
            assert all(arc is None or arc.chain.isValid() for arc in solvedArcsCopy)

        return Frame(self.row, self.commit, solvedArcsCopy, openArcsCopy, self.lastArc)

    @staticmethod
    def reserveArcListCapacity(theList, newLength):
        if len(theList) >= newLength:
            return
        for _ in range(newLength - len(theList)):
            theList.append(None)

    @staticmethod
    def cleanUpArcList(theList: list[Arc | None], olderThanRow: BatchRow, alsoTrimBack: bool = True):
        # Remove references to arcs that were closed earlier than `olderThanRow`
        for j, arc in enumerate(theList):
            if arc and 0 <= arc.closedAt < olderThanRow:
                theList[j] = None

        # Cull None items at the end of the list
        if alsoTrimBack:
            while theList and not theList[-1]:
                del theList[-1]

        return theList

    def flattenLanes(self, hiddenCommits: Set[Oid]) -> tuple[list[tuple[int, int]], int]:
        """Flatten the lanes so there are no unused columns in-between the lanes."""

        row = int(self.row)

        # Filter arcs

        def genArcsAbove():
            gen1 = (arc for arc in self.solvedArcs if arc)
            gen2 = (arc for arc in self.openArcs if arc and arc.openedAt < row)
            if hiddenCommits:
                gen1 = (arc for arc in gen1 if arc.isVisible(hiddenCommits, row))
                gen2 = (arc for arc in gen2 if arc.isVisible(hiddenCommits, row))
            yield from gen1
            yield from gen2

        def genArcsBelow():
            gen = (arc for arc in self.openArcs if arc)
            if hiddenCommits:
                gen = (arc for arc in gen if arc.isVisible(hiddenCommits, row, BatchRow.__le__))
            yield from gen

        # Sort arcs by Chain Birth Row

        def sortArc(a: Arc):
            return (int(a.chain.topRow) << 16) + a.lane

        arcsAbove = sorted(genArcsAbove(), key=sortArc, reverse=True)
        arcsBelow = sorted(genArcsBelow(), key=sortArc, reverse=True)

        # Assign columns to all lanes used by arcs above and below

        N = max(len(self.solvedArcs), len(self.openArcs))
        mapAbove = [-1] * N
        mapBelow = [-1] * N

        column = -1
        while arcsAbove or arcsBelow:
            column += 1

            if arcsAbove:
                a = arcsAbove.pop()
                mapAbove[a.lane] = column

            if arcsBelow:
                a = arcsBelow.pop()
                mapBelow[a.lane] = column

        remapTable = list(zip(mapAbove, mapBelow, strict=True))
        return remapTable, column + 1


class PlaybackState(Frame):
    def __init__(self, keyframe: Frame):
        super().__init__(
            row=keyframe.row,
            commit=keyframe.commit,
            solvedArcs=[],
            openArcs=[],
            lastArc=keyframe.lastArc)

        self.solvedArcs = keyframe.solvedArcs.copy()
        self.openArcs = keyframe.openArcs.copy()
        self.callingNextWillAdvanceFrame = True
        self.seenCommits: set[Oid] = set()

    def advanceToNextRow(self):
        if not self.callingNextWillAdvanceFrame:
            self.callingNextWillAdvanceFrame = True
            return

        # Put commit from previous iteration into set of previously-seen commits
        if self.commit:
            self.seenCommits.add(self.commit)

        goalFound = False
        goalRow = BATCHROW_UNDEF
        goalCommit = None

        while self.lastArc.nextArc:
            arc: Arc = self.lastArc.nextArc

            if not goalFound and arc.openedAt > self.row:
                # The goal row is determined by the first arc opened at a row greater than the player's current row.
                goalFound = True
                goalRow = arc.openedAt
                goalCommit = arc.openedBy
            elif goalFound and arc.openedAt > goalRow:
                # If we went past the goal row, we have seen all arcs opened at the goal row. Stop.
                break
            else:
                # Keep iterating on arcs in the goal row. Gather them in a frame.
                pass

            self.reserveArcListCapacity(self.solvedArcs, arc.lane + 1)
            self.reserveArcListCapacity(self.openArcs, arc.lane + 1)

            self.solvedArcs[arc.lane] = self.openArcs[arc.lane]  # move any open arc to "just closed"
            self.openArcs[arc.lane] = arc
            self.lastArc = arc

        if not goalFound:
            raise StopIteration()

        assert self.row < goalRow
        assert goalCommit is not None

        self.row = goalRow
        self.commit = goalCommit

        assert isinstance(self.row, BatchRow)

    def __iter__(self):
        return self

    def __next__(self):
        self.advanceToNextRow()
        return self


class Graph:
    keyframes: list[Frame]
    keyframeRows: list[BatchRow]

    startArc: Arc
    """
    Start sentinel of the linked list of Arcs. Guaranteed to never be None.
    Use startArc.nextArc to get to the first actual arc.
    """

    commitRows: dict[Oid, BatchRow]
    ownBatches: list[int]

    volatilePlayer: PlaybackState | None

    def __init__(self):
        self.keyframes = []
        self.keyframeRows = []
        self.commitRows = {}
        self.startArc = Arc(
            openedAt=BATCHROW_UNDEF,
            closedAt=BATCHROW_UNDEF,
            chain=ChainHandle(BATCHROW_UNDEF, BATCHROW_UNDEF),
            lane=-1,
            openedBy=None,
            closedBy=None,
            junctions=[],
            nextArc=None)
        self.ownBatches = []
        self.volatilePlayer = None

    def __del__(self):
        self.freeOwnBatches()

    def freeOwnBatches(self):
        for b in self.ownBatches:
            BatchRow.BatchManager.freeBatch(b)
        self.ownBatches = []

    def shallowCopyFrom(self, source: Graph):
        assert not set(self.ownBatches).intersection(set(source.ownBatches))

        # Free up owned batches
        self.freeOwnBatches()

        self.keyframes = source.keyframes
        self.keyframeRows = source.keyframeRows
        self.startArc = source.startArc
        self.commitRows = source.commitRows
        self.ownBatches = source.ownBatches

        source.ownBatches = []

        source.volatilePlayer = None
        self.volatilePlayer = None

    def isEmpty(self):
        return self.startArc.nextArc is None

    def getCommitRow(self, oid: Oid):
        return int(self.commitRows[oid])

    def saveKeyframe(self, frame: Frame) -> int:
        assert len(self.keyframes) == len(self.keyframeRows)

        kfID = bisect.bisect_left(self.keyframeRows, frame.row)
        if kfID < len(self.keyframes) and self.keyframes[kfID].row == frame.row:
            logger.info(f"Not overwriting existing keyframe {kfID}")
            assert self.keyframes[kfID] == frame.sealCopy()
        else:
            kf = frame.sealCopy()
            self.keyframes.insert(kfID, kf)
            self.keyframeRows.insert(kfID, frame.row)
        return kfID

    def getBestKeyframeID(self, row: int) -> int:
        """
        Attempts to find a keyframe closest to `row` in the frame sequence.
        If an adequate keyframe is found, return its index into the keyframes list; otherwise, return -1.
        Note that the returned value is an **index into the list of keyframes**; it is NOT a frame row number.

        If a valid keyframe was found, its row is guaranteed to be lower or equal to `row`.
        This function never returns a keyframe located at a greater row than `row`.

        If the keyframe occurs before the desired `row`, you can create a PlaybackState from that keyframe
        and iterate the PlaybackState until it reaches the desired row.
        """
        assert row >= 0
        assert len(self.keyframes) == len(self.keyframeRows)

        bestKeyframeID = bisect.bisect_right(self.keyframeRows, row) - 1
        if bestKeyframeID < 0:
            return -1

        bestKeyframeRow = self.keyframeRows[bestKeyframeID]
        assert 0 <= bestKeyframeRow <= row

        return bestKeyframeID

    def startPlayback(self, goalRow: int = 0, oneOff: bool = False) -> PlaybackState:
        kfID = self.getBestKeyframeID(goalRow)
        if kfID >= 0:
            kf = self.keyframes[kfID]
        else:
            kf = self.initialKeyframe()

        if oneOff and self.volatilePlayer and goalRow >= self.volatilePlayer.row >= kf.row:
            player = self.volatilePlayer
        else:
            player = PlaybackState(kf)

        # Position playback context on target row
        try:
            volatileKeyframeCounter = 1
            assert player.row <= goalRow, f"{player.row} {goalRow}"
            while player.row < goalRow:
                player.advanceToNextRow()  # raises StopIteration if depleted

                # Save keyframes every now and then
                if player.row - kf.row >= volatileKeyframeCounter:
                    volatileKeyframeCounter *= 2
                    self.saveKeyframe(player)

            assert player.row == goalRow
            player.callingNextWillAdvanceFrame = False  # let us re-obtain current frame by calling next()
        except StopIteration:
            # Depleted - make sure we get StopIteration next time we call `next`.
            assert player.callingNextWillAdvanceFrame
            assert player.lastArc.nextArc is None

        if oneOff:
            self.volatilePlayer = player

        return player

    def getCommitFrame(self, commit: Oid, unsafe=False) -> Frame:
        row = self.getCommitRow(commit)
        return self.getFrame(row, unsafe)

    def getFrame(self, row: int = 0, unsafe=False) -> Frame:
        assert row >= 0

        kfID = self.getBestKeyframeID(row)

        if kfID >= 0 and self.keyframes[kfID].row == row:
            # Cache hit
            frame = self.keyframes[kfID]
        else:
            # Cache miss
            frame = self.startPlayback(row)
            if not unsafe:
                frame = frame.sealCopy()

        assert frame.row == row, f"frame({int(frame.row)})/row({row}) mismatch"
        return frame

    def initialKeyframe(self):
        return Frame(
            row=BATCHROW_UNDEF,
            commit=self.startArc.openedBy,
            solvedArcs=[],
            openArcs=[],
            lastArc=self.startArc)

    def deleteKeyframesDependingOnRowsAbove(self, row: int):
        """
        Deletes all keyframes containing any arcs opened above the given row.
        """

        if len(self.keyframes) == 0:
            return

        # Get a starting keyframe for a row <= the desired row.
        kfID = self.getBestKeyframeID(row)
        if kfID < 0:
            kfID = 0

        # Fast-forward to the next keyframe with a row >= the desired row.
        while kfID < len(self.keyframes) and self.keyframes[kfID].row < row:
            kfID += 1

        # Once we reach keyframes occuring at a row >= `row`, see if we should stop deleting keyframes.
        # All open arcs, and all non-stale solved arcs, must be independent of rows <= `row`.
        while kfID < len(self.keyframes):
            kf = self.keyframes[kfID]

            if (all(arc is None or arc.isIndependentOfRowsAbove(row) for arc in kf.openArcs)
                    and all(arc is None or arc.isIndependentOfRowsAbove(row) or arc.isStale(row) for arc in kf.solvedArcs)):
                break
            else:
                kfID += 1

        # Delete the keyframes up to kfID.
        self.keyframes = self.keyframes[kfID:]
        self.keyframeRows = self.keyframeRows[kfID:]
        assert len(self.keyframes) == len(self.keyframeRows)

    def deleteArcsDependingOnRowsAbove(self, row: int):
        """
        Deletes all arcs opened before the given row.
        """

        if row == 0:
            return

        # In debug mode, bulldoze opening commits in dead arcs so they stand out in the debugger (make them dangling)
        if APP_DEBUG and self.startArc.nextArc:
            for deadArc in self.startArc.nextArc:
                if deadArc.openedAt >= row:
                    break
                deadArc.openedAt = BATCHROW_UNDEF
                deadArc.openedBy = DEAD_VALUE

        # Rewire top of list
        self.startArc.nextArc =\
            next((arc for arc in self.startArc if arc.openedAt >= row), None)

    def insertFront(self, frontGraph: Graph, numRowsToInsert: int):
        """
        Inserts contents of frontGraph at the beginning of this graph.
        (This function is invoked as step 2 of merging two graphs.)
        """

        # Graph to insert is empty? Bail.
        if frontGraph.startArc.nextArc is None:
            return

        # Don't want to insert any rows? Bail.
        if numRowsToInsert == 0:
            return

        # Find out the last arc we want to take from the front graph
        theirLastArc = next(arc for arc in frontGraph.startArc
                            if arc.nextArc is None or arc.nextArc.openedAt >= numRowsToInsert)
        assert theirLastArc is not None
        assert theirLastArc != frontGraph.startArc

        # In the arc linked list, rewire their last arc onto my first arc.
        theirLastArc.nextArc = self.startArc.nextArc

        # Rewire my top sentinel onto their first actual arc
        self.startArc.nextArc = frontGraph.startArc.nextArc

        # Steal their keyframes
        lastFrontKeyframeID = frontGraph.getBestKeyframeID(numRowsToInsert - 1)
        if lastFrontKeyframeID >= 0:
            assert len(self.keyframes) == len(self.keyframeRows)
            assert len(frontGraph.keyframes) == len(frontGraph.keyframeRows)
            self.keyframes = frontGraph.keyframes[:lastFrontKeyframeID + 1] + self.keyframes
            self.keyframeRows = frontGraph.keyframeRows[:lastFrontKeyframeID + 1] + self.keyframeRows

    def testConsistency(self):
        """ Very expensive consistency check for unit testing """

        # Verify chains
        if self.startArc.nextArc:
            for a in self.startArc.nextArc:
                assert a.chain.isValid()

        # Verify keyframes
        playback = self.startPlayback(0)
        for row, keyframe in zip(self.keyframeRows, self.keyframes, strict=True):
            while playback.commit != keyframe.commit:
                next(playback)
            frame1 = playback.sealCopy()
            frame2 = keyframe.sealCopy()
            assert frame1 == frame2, f"Keyframe at row {row} doesn't match actual graph state"



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import QValidator


class ReplaceSpacesWithDashes(QValidator):
    def validate(self, text: str, pos: int) -> tuple[QValidator.State, str, int]:
        text = text.replace(" ", "-")
        return QValidator.State.Acceptable, text, pos



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.toolbox import *


def makeBrandedDialogLayout(
        dialog: QDialog,
        titleText: str,
        subtitleText: str = "",
        multilineSubtitle: bool = False
):
    gridLayout = QGridLayout(dialog)

    iconLabel = QLabel(dialog)
    iconLabel.setMaximumSize(QSize(56, 56))
    iconLabel.setPixmap(QPixmap("assets:icons/gitfourchette"))
    iconLabel.setScaledContents(True)
    iconLabel.setMargin(8)

    horizontalSpacer = QSpacerItem(0, 1, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)

    titleLayout = QVBoxLayout()
    titleLayout.setSpacing(0)
    titleLayout.setContentsMargins(0, 0, 0, 0)
    title = QLabel(titleText, dialog)
    title.setTextFormat(Qt.TextFormat.RichText)
    title.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
    tweakWidgetFont(title, 150, bold=True)
    titleLayout.addWidget(title)

    if subtitleText:
        title.setAlignment(Qt.AlignmentFlag.AlignBottom)
        subtitleWidgets = []

        if multilineSubtitle:
            subtitle = QLabel(subtitleText)
            subtitle.setWordWrap(True)
            subtitleWidgets.append(subtitle)
        else:
            for line in subtitleText.splitlines():
                subtitleWidgets.append(QElidedLabel(line, dialog))

        for subtitle in subtitleWidgets:
            subtitle.setAlignment(Qt.AlignmentFlag.AlignTop)
            tweakWidgetFont(subtitle, relativeSize=90)
            titleLayout.addWidget(subtitle)

    gridLayout.addWidget(iconLabel, 0, 0, 1, 1)
    gridLayout.addItem(horizontalSpacer, 0, 1, 1, 1)
    gridLayout.addLayout(titleLayout, 0, 3, 1, 1)

    if subtitleText:
        breather = QSpacerItem(0, 8, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        gridLayout.addItem(breather, 1, 0)

    return gridLayout


def convertToBrandedDialog(
        dialog: QDialog,
        promptText: str = "",
        subtitleText: str = "",
        multilineSubtitle: bool = False,
):
    if not promptText:
        promptText = escape(dialog.windowTitle())

    innerContent = QWidget(dialog)
    innerContent.setLayout(dialog.layout())
    innerContent.layout().setContentsMargins(0, 0, 0, 0)

    gridLayout = makeBrandedDialogLayout(dialog, promptText, subtitleText, multilineSubtitle)
    gridLayout.addWidget(innerContent, 2, 3, 1, 1)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import colors
from gitfourchette.localization import _
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class GpgButton(QToolButton):
    def __init__(self, parent):
        super().__init__(parent)

        self.gpgFlag = False
        self.gpgKey = ""
        self.willSign = False

        self.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

    def setup(self, flag: bool, key: str):
        self.gpgFlag = flag
        self.gpgKey = key
        self.willSign = bool(flag and key)
        canSign = bool(key)

        keyAction = QAction(self)
        if key:
            keyAction.setText(_("Signing key: {0}", escamp(key)))
        else:
            keyAction.setText(_("Signing key not configured ({0})", "user.signingKey"))

        keyAction.setEnabled(False)

        toggleAction = QAction(self)
        if canSign:
            toggleAction.setText(_("Enable &GPG-signing for this commit"))
        else:
            toggleAction.setText(_("This commit will not be GPG-signed."))
        toggleAction.setEnabled(canSign)
        toggleAction.setCheckable(canSign)
        toggleAction.setChecked(self.willSign)
        toggleAction.toggled.connect(lambda t: self.updateConfig(t))

        self.addActions([toggleAction, keyAction])

        self.updateConfig(self.willSign)

    def updateConfig(self, willSign: bool):
        self.willSign = willSign

        flag = self.gpgFlag
        key = self.gpgKey
        lines = []

        if willSign:
            lines.append(tagify(_("This commit [will be GPG-signed] with your key {yourkey} (configured in {keyconf})."), "<b>"))
        else:
            lines.append(_("This commit will not be GPG-signed."))

        if willSign:
            pass
        elif key:
            lines.append(_("Click this button to GPG-sign the commit with your key {yourkey} (configured in {keyconf})."))
        else:
            lines.append(_("Your GPG-signing key isn’t configured in {keyconf}."))

        if not key:
            pass
        elif flag:
            lines.append(_("This repository is configured to GPG-sign commits automatically ({flagconf})."))
        else:
            lines.append(_("Tip: You can configure {flagconf} if you want to GPG-sign every commit."))

        tip = "<html style='white-space: pre-wrap;'>" + paragraphs(*lines)
        tip = tip.format(flagconf="<i>commit.gpgSign</i>",
                         keyconf="<i>user.signingKey</i>",
                         yourkey=hquo(key))

        if willSign:
            icon = stockIcon("gpg-key", f"gray={colors.olive.name()}")
        else:
            icon = stockIcon("gpg-key-fail")

        self.setIcon(icon)
        self.setToolTip(tip)

    def explicitSign(self):
        return self.willSign

    def explicitNoSign(self):
        return not self.willSign



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_newbranchdialog import Ui_NewBranchDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class NewBranchDialog(QDialog):
    def __init__(
            self,
            initialName: str,
            target: str,
            targetSubtitle: str,
            upstreams: list[str],
            reservedNames: list[str],
            allowSwitching: bool,
            showSubmoduleControls: bool,
            parent=None):

        super().__init__(parent)

        self.ui = Ui_NewBranchDialog()
        self.ui.setupUi(self)

        self.ui.nameEdit.setText(initialName)
        self.acceptButton.setText(_("&Create"))

        self.ui.upstreamComboBox.addItems(upstreams)

        # hack to trickle down initial 'toggled' signal to combobox
        self.ui.upstreamCheckBox.setChecked(True)
        self.ui.upstreamCheckBox.setChecked(False)

        if not upstreams:
            self.ui.upstreamCheckBox.setChecked(False)
            self.ui.upstreamCheckBox.setVisible(False)
            self.ui.upstreamComboBox.setVisible(False)

        if not allowSwitching:
            switchCheckBox = self.ui.switchToBranchCheckBox
            switchCheckBox.setEnabled(False)
            switchCheckBox.setChecked(False)
            switchCheckBox.setText(switchCheckBox.text() + "\n" + _("(blocked by conflicts)"))
            self.ui.recurseSubmodulesCheckBox.setChecked(False)

        if not showSubmoduleControls:
            self.ui.recurseSubmodulesCheckBox.setChecked(False)
            self.ui.recurseSubmodulesCheckBox.setVisible(False)

        nameTaken = _("This name is already taken by another local branch.")
        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(self.acceptButton)
        validator.connectInput(self.ui.nameEdit, lambda name: nameValidationMessage(name, reservedNames, nameTaken))
        validator.run()

        convertToBrandedDialog(self, _("New branch"),
                               _("Commit at tip:") + " " + target + "\n" + tquo(targetSubtitle))

        self.ui.nameEdit.setFocus()
        self.ui.nameEdit.selectAll()
        self.ui.nameEdit.setValidator(ReplaceSpacesWithDashes())
        self.preSelectLeaf(self.ui.nameEdit)

        packDialog(self, lockHeight=True)

    @property
    def acceptButton(self):
        return self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

    @staticmethod
    def preSelectLeaf(nameEdit: QLineEdit):
        """
        Pre-select leaf name (e.g. in "folder/leaf" select only "leaf")
        """
        text = nameEdit.text()
        try:
            afterSlash = text.rindex("/") + 1
            nameEdit.setSelection(afterSlash, len(text) - afterSlash)
        except ValueError:
            pass



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import urllib

from gitfourchette.toolbox import *


HTTPS_PORT = 443


@dataclasses.dataclass
class WebHost:
    name: str
    branchPrefix: str
    port: int = HTTPS_PORT

    @staticmethod
    def makeLink(remoteUrl: str, branch: str = ""):
        host, path = splitRemoteUrl(remoteUrl)

        if not host:
            return "", ""

        path = path.removesuffix(".git")

        try:
            hostInfo = WEB_HOSTS[host]
            hostName = hostInfo.name
        except KeyError:
            hostInfo = WEB_HOSTS["github.com"]  # fall back to GitHub's scheme
            hostName = host

        port = "" if hostInfo.port == HTTPS_PORT else f":{hostInfo.port}"
        suffix = ""

        if branch:
            suffix = hostInfo.branchPrefix + urllib.parse.quote(branch, safe='/')

        return f"https://{host}{port}/{path}{suffix}", hostName


WEB_HOSTS = {
    "github.com": WebHost("GitHub", "/tree/"),
    "gitlab.com": WebHost("GitLab", "/-/tree/"),
    "git.sr.ht": WebHost("Sourcehut", "/tree/"),
    "codeberg.org": WebHost("Codeberg", "/src/branch/"),
    "git.launchpad.net": WebHost("Launchpad", "/log?h="),
    "bitbucket.org": WebHost("Bitbucket", "/src/"),
}



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import enum
from typing import ClassVar

from gitfourchette.gitdriver import GitDeltaSource
from gitfourchette.porcelain import NULL_OID, Oid
from gitfourchette.qt import *
from gitfourchette.toolbox import *

if TYPE_CHECKING:
    from gitfourchette.graphview.commitlogmodel import SpecialRow

PUSH_INTERVAL_MS = 500


class NavFlags(enum.IntFlag):
    AllowWriteIndex = enum.auto()
    "Allow writing to the index while loading the diff at this location."

    AllowLargeFiles = enum.auto()
    "Bypass file size and line length limits to display the diff at this location."

    ForceDiff = enum.auto()
    "Force reloading the pygit2.Diff of the commit or workdir."

    ForceRecreateDocument = enum.auto()
    "Force re-rendering the DiffDocument."

    BypassFileSelect = enum.auto()
    "Bypass automatic selection in FileList when loading this location."

    BypassCommitSelect = enum.auto()
    "Bypass automatic selection in GraphView when loading this location."

    ActivateWindow = enum.auto()
    "Bring window to the foreground after jumping to this location."

    FuzzyPath = enum.auto()
    "The path may contain wildcards and should be expanded with fnmatch."

    DefaultFlags = 0

    KeepFlagsOnRefresh = AllowWriteIndex | AllowLargeFiles
    """
    Flags that should be honored again by the Jump task after refreshing the
    RepoWidget on the same NavLocator.

    All other flags are "one-time" flags: they are meant to be consumed by the
    Jump task, then they are discarded before the NavLocator is installed
    as the RepoWidget's current locator.

    Note that NavHistory always strips *all* flags when saving a NavLocator.
    """


@enum.unique
class NavContext(enum.IntEnum):
    """
    State of a patch in the staging pipeline
    """

    EMPTY       = 0
    COMMITTED   = 1
    WORKDIR     = 2
    UNSTAGED    = 3
    STAGED      = 4
    SPECIAL     = 5

    def isWorkdir(self) -> bool:
        return self == NavContext.WORKDIR or self == NavContext.UNSTAGED or self == NavContext.STAGED

    @classmethod
    def fromGitDeltaSource(cls, source: GitDeltaSource):
        _deltaContextToNavContext = {
            GitDeltaSource.Dirty: NavContext.UNSTAGED,
            GitDeltaSource.Index: NavContext.STAGED,
            GitDeltaSource.Commit: NavContext.COMMITTED,
        }
        return _deltaContextToNavContext[source]


@dataclasses.dataclass(frozen=True)
class NavLocator:
    """
    Resource locator within a repository.
    Used to navigate the UI to a specific area of a repository.
    """

    Empty: ClassVar[NavLocator]

    context: NavContext = NavContext.EMPTY
    commit: Oid = NULL_OID
    path: str = ""
    ref: str = ""

    cursorChar: int = 0
    """
    Character position of the text cursor.
    """

    cursorLine: int = 0
    """
    Line number of the text cursor. CodeView will attempt to keep the cursor on
    this line in case of a mismatch with `cursorChar` (which may occur when file
    contents change in the workdir).
    """

    scrollChar: int = 0
    """
    Character position of the top-left corner in the scrollable viewport.
    """

    flags: NavFlags = NavFlags.DefaultFlags
    """
    Transient flags, not saved in NavHistory.
    """

    selectedCommits: tuple[Oid, ...] = ()
    """
    Currently selected commits in GraphView. Can be used to diff two commits.
    If empty and `context` is COMMITTED, `commit` is assumed to be the sole
    selected commit.
    This is a tuple to enforce immutability.
    """

    URL_AUTHORITY: ClassVar[str] = "jump"

    if APP_DEBUG:
        def __post_init__(self):
            assert not self.path.endswith("/")
            assert isinstance(self.context, NavContext)
            assert isinstance(self.commit, Oid)
            assert isinstance(self.path, str)
            assert isinstance(self.ref, str)
            assert isinstance(self.selectedCommits, tuple)
            hasCommit = self.commit != NULL_OID
            hasRef = bool(self.ref)
            if self.context == NavContext.COMMITTED:
                assert hasCommit or hasRef
                assert hasCommit ^ hasRef
                assert not hasRef or self.ref == "HEAD" or self.ref.startswith("refs/")
            elif self.context == NavContext.SPECIAL:
                assert self.path, "special locator requires 'path' field (a SpecialRow value)"
                assert not hasRef
            else:
                assert not hasCommit
                assert not hasRef

    def __bool__(self):
        """
        Return True if the locator's context is anything but EMPTY.

        The locator is NOT considered empty when the path is empty but the context isn't
        (e.g. in the STAGED context, with no files selected.)
        """
        return self.context.value != NavContext.EMPTY

    def __repr__(self) -> str:
        return F"{self.__class__.__name__}({self.contextKey:.8} {self.path})"

    @staticmethod
    def inCommit(oid: Oid, path: str = "") -> NavLocator:
        return NavLocator(context=NavContext.COMMITTED, commit=oid, path=path)

    @staticmethod
    def inRef(ref: str, path: str = "") -> NavLocator:
        return NavLocator(context=NavContext.COMMITTED, ref=ref, path=path)

    @staticmethod
    def inUnstaged(path: str = "") -> NavLocator:
        return NavLocator(context=NavContext.UNSTAGED, path=path)

    @staticmethod
    def inStaged(path: str = "") -> NavLocator:
        return NavLocator(context=NavContext.STAGED, path=path)

    @staticmethod
    def inWorkdir(path: str = "") -> NavLocator:
        return NavLocator(context=NavContext.WORKDIR, path=path)

    @staticmethod
    def inSpecial(special: SpecialRow) -> NavLocator:
        return NavLocator(context=NavContext.SPECIAL, path=str(special))

    def isSimilarEnoughTo(
            self,
            other: NavLocator,
            considerPaths: bool = True,
    ) -> bool:
        """
        Coarse equality: Compare context, commit(s) & path.
        Ignores flags & position in diff.
        """

        contextA = self.context
        contextB = other.context

        if considerPaths:
            pass
        elif contextA == NavContext.SPECIAL:
            # Special locators store their kind in the path field, so it cannot be ignored
            considerPaths = True
        else:
            # Round down unstaged/staged to workdir
            contextA = NavContext.WORKDIR if contextA.isWorkdir() else contextA
            contextB = NavContext.WORKDIR if contextB.isWorkdir() else contextB

        return (
            contextA == contextB
            and self.commit == other.commit
            and self.selectedCommits == other.selectedCommits
            and self.ref == other.ref
            and (not considerPaths or self.path == other.path)
        )

    def hasFlags(self, flags: NavFlags):
        return flags == (self.flags & flags)

    def url(self):
        if self.context == NavContext.COMMITTED:
            fragment = str(self.commit)
        else:
            fragment = self.context.name

        query = {}
        if self.flags != NavFlags.DefaultFlags:
            query["flags"] = str(self.flags.value)

        return makeInternalLink(NavLocator.URL_AUTHORITY, self.path, fragment, **query)

    def replace(self, **kwargs) -> NavLocator:
        return dataclasses.replace(self, **kwargs)

    def coarse(self, keepFlags=False) -> NavLocator:
        flags = NavFlags.DefaultFlags if not keepFlags else self.flags
        return NavLocator(
            context=self.context,
            commit=self.commit,
            path=self.path,
            ref=self.ref,
            flags=flags,
            selectedCommits=self.selectedCommits,
        )

    def withExtraFlags(self, flags: NavFlags) -> NavLocator:
        return self.replace(flags=self.flags | flags)

    def withoutFlags(self, flags: NavFlags) -> NavLocator:
        return self.replace(flags=self.flags & ~flags)

    def commitDiffAB(self) -> tuple[Oid, Oid] | None:
        """
        When 2 commits are selected, return the pair of commit oids forming an
        A/B diff. Return None otherwise.
        """
        if self.context != NavContext.COMMITTED or len(self.selectedCommits) != 2:
            return None

        assert isinstance(self.selectedCommits, tuple)
        assert self.commit in self.selectedCommits
        return self.selectedCommits

    def swapABCommits(self) -> NavLocator:
        """
        When 2 commits are selected, return a new locator where the A and B
        commits are swapped. The "current" commit remains unchanged.
        Do not call if this locator doesn't represent an A/B commit diff!
        """
        if not self.commitDiffAB():
            raise ValueError("locator does not represent an A/B commit diff")

        selected = tuple(reversed(self.selectedCommits))
        return self.replace(selectedCommits=selected)

    @staticmethod
    def parseUrl(url: QUrl):
        assert url.authority() == NavLocator.URL_AUTHORITY
        assert url.hasFragment()

        frag = url.fragment()
        path = url.path()

        if path:  # fix up non-empty path
            assert path.startswith("/")
            path = path.removeprefix("/")

        try:
            context = NavContext[frag]
            commit = NULL_OID
        except KeyError:
            context = NavContext.COMMITTED
            commit = Oid(hex=frag)

        flags = NavFlags.DefaultFlags
        query = QUrlQuery(url.query())
        if not query.isEmpty():
            strFlags = query.queryItemValue("flags")
            flags = NavFlags(int(strFlags))

        return NavLocator(context, commit, path, flags=flags)

    @property
    def contextKey(self):
        if self.context == NavContext.SPECIAL:
            return f"@special@{self.path}"
        elif self.context != NavContext.COMMITTED:
            return self.context.name
        elif self.ref:
            return self.ref
        elif not self.selectedCommits:
            return str(self.commit)
        else:
            return ";".join(str(c) for c in self.selectedCommits)

    @property
    def fileKey(self):
        """ For NavHistory.recallFileInContext(). """
        return f"{self.contextKey}:{self.path}"

    @property
    def hash7(self):
        """ Return the first seven hex digits of the commit hash (unit testing helper) """
        return str(self.commit)[:7]

NavLocator.Empty = NavLocator()


class NavHistory:
    """
    History of the files that the user has viewed in a repository's commit log and workdir.
    """

    history: list[NavLocator]
    "Stack of position snapshots."

    current: int
    "Current position in the history stack. Hitting next/forward moves this index."

    recent: dict[str, NavLocator]
    "Most recent NavPos by context key"

    lastPushTime: int
    """Millisecond timestamp of the last modification to the history,
    to avoid pushing a million entries when dragging the mouse, etc."""

    def __init__(self):
        self.history = []
        self.recent = {}
        self.current = 0
        self.lastPushTime = 0
        self.ignoreDelay = False

        # In a real use case, locators are dropped from the history if push()
        # calls occur in quick succession. This avoids polluting the history
        # with unimportant entries when the user drags the mouse across
        # GraphView, for instance. However, in unit tests, navigation occurs
        # blazingly quickly, but we want each location to be recorded in the
        # history.
        self.ignoreDelay |= APP_TESTMODE

    def copy(self) -> NavHistory:
        copy = NavHistory()
        copy.history = self.history.copy()
        copy.current = self.current
        copy.recent = self.recent.copy()
        copy.lastPushTime = self.lastPushTime
        return copy

    def push(self, pos: NavLocator):
        if not pos:
            return

        # Clear volatile flags
        if pos.flags != NavFlags.DefaultFlags:
            pos = pos.replace(flags=NavFlags.DefaultFlags)

        self.recent[pos.contextKey] = pos
        self.recent[pos.fileKey] = pos

        if pos.context in [NavContext.STAGED, NavContext.UNSTAGED]:
            # This is a "concrete" workdir locator. Redirect to it when jumping to the
            # "abstract" workdir locator (e.g. by clicking on Uncommitted Changes).
            self.recent[NavContext.WORKDIR.name] = pos

        now = QDateTime.currentMSecsSinceEpoch()
        if self.ignoreDelay:
            recentPush = False
        else:
            recentPush = (now - self.lastPushTime) < PUSH_INTERVAL_MS

        if len(self.history) > 0 and \
                (recentPush or self.history[self.current].isSimilarEnoughTo(pos)):
            # Update in-place; don't update lastPush timestamp
            self.history[self.current] = pos
        else:
            if self.current < len(self.history) - 1:
                self.trimFuture()
            self.history.append(pos)
            self.current = len(self.history) - 1
            self.lastPushTime = now

    def trimFuture(self):
        self.history = self.history[: self.current + 1]
        assert not self.canGoForward()

    def refine(self, locator: NavLocator):
        """
        Attempt to make a locator more precise by looking up the most recent
        matching locator in the history.

        For example, given a locator with context=UNSTAGED but without a path,
        refine() will return the locator for an unstaged file that was most
        recently saved to the history.

        In addition, diff cursor/scroll positions will be filled in if the
        history contains them.

        If refining isn't possible, this function returns the same locator as
        the input.
        """

        # Never refine special locators
        if locator.context == NavContext.SPECIAL:
            return locator

        originalLocator = locator

        # If no path is specified, attempt to recall any path in the same context
        if not locator.path:
            locator = self.recent.get(locator.contextKey, NavLocator.Empty)
            if not locator:
                locator = originalLocator

        locator = self.recent.get(locator.fileKey, NavLocator.Empty) or locator

        # Restore volatile flags
        if originalLocator.flags != locator.flags:
            locator = locator.replace(flags=originalLocator.flags)

        return locator

    def canGoForward(self):
        count = len(self.history)
        return count > 0 and self.current < count - 1

    def canGoBack(self):
        count = len(self.history)
        return count > 0 and self.current > 0

    def canGoDelta(self, delta: int):
        assert delta == -1 or delta == 1
        if delta > 0:
            return self.canGoForward()
        else:
            return self.canGoBack()

    def navigateBack(self):
        if not self.canGoBack():
            return None
        self.current -= 1
        return self.history[self.current]

    def navigateForward(self):
        if not self.canGoForward():
            return None
        self.current += 1
        return self.history[self.current]

    def navigateDelta(self, delta: int):
        assert delta == -1 or delta == 1
        if delta > 0:
            return self.navigateForward()
        else:
            return self.navigateBack()

    def popCurrent(self):
        if self.current < len(self.history):
            return self.history.pop(self.current)
        else:
            return None

    def getTextLog(self):  # pragma: no cover
        s = "------------ NAV LOG ------------"
        i = len(self.history) - 1
        for h in reversed(self.history):
            s += "\n"
            if i == self.current:
                s += "---> "
            else:
                s += "     "
            s += f"{h.contextKey[:7]} {h.path:32} {h.scrollChar} {h.cursorChar}"
            i -= 1
        return s



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

from gitfourchette.porcelain import Oid
from gitfourchette.qt import *
from gitfourchette.toolbox.benchmark import benchmark
from gitfourchette.toolbox.textutils import qstringLength

if TYPE_CHECKING:
    from pygments.lexer import Lexer
    from pygments.token import _TokenType

    type LineTokenization = list[tuple[_TokenType, int]]
    "Type alias for the tokens in a line + their UTF-16 lengths."

_EmptyLineTokenization: LineTokenization = []


class LexJob(QObject):
    type KeyType = str | Oid

    ChunkSize = 5000  # tokens
    ScheduleInitialDelay = 0  # ms
    ScheduleInterval = 0  # ms
    MaxLowQualityLines = 100
    MaxLowQualityLineLength = 200

    pulse = Signal()

    def __init__(self, lexer: Lexer, data: bytes | str, fileKey: KeyType):
        # Don't bind the QObject to a parent to allow Python's refcounting to
        # purge evicted cache entries that are not currently in use by the UI.
        super().__init__(None)
        self.setObjectName("LexJob")

        assert fileKey
        assert data, "don't create a LexJob without some data"

        self.lexer = lexer
        self.lqTokenMap: dict[int, LineTokenization] = {}
        self.hqTokenMap: dict[int, LineTokenization] = {1: []}
        self.fileKey = fileKey
        self.fileSize = len(data)

        self.currentLine = 1
        self.lexGen = lexer.get_tokens(data)  # type: ignore[arg-type] # pygments stubs unaware that data can be bytes

        self.scheduler = QTimer(self)
        self.scheduler.setSingleShot(True)
        self.scheduler.timeout.connect(self.lexChunk)

        self.requestedLine = 0
        assert not self.lexingComplete

    @property
    def lexingComplete(self):
        return self.currentLine == 0

    def tokens(self, lineNumber: int, fallbackText: str) -> LineTokenization:
        if self.lexingComplete or self.currentLine > lineNumber:
            return self.hqTokenMap[lineNumber]

        # Lex job hasn't reached this line yet.
        # Schedule high-quality lexing up to this line.
        if self.requestedLine < lineNumber:
            self.requestedLine = lineNumber
            self.start()  # Initiate chunking

        # Fall back to low-quality lexing on this line
        # to minimize flashing while the job is busy.
        try:
            lqTokens = self.lqTokenMap[lineNumber]
        except KeyError:
            if (len(self.lqTokenMap) > LexJob.MaxLowQualityLines
                    or len(fallbackText) > LexJob.MaxLowQualityLineLength):
                # To ease CPU load, skip long lines and stop LQ-lexing "below the fold".
                lqTokens = _EmptyLineTokenization
            else:
                # Perform low-quality lexing and cache the result.
                lqTokens = [(t, qstringLength(v)) for _i, t, v in self.lexer.get_tokens_unprocessed(fallbackText)]
                self.lqTokenMap[lineNumber] = lqTokens
        return lqTokens

    def start(self):
        assert not self.lexingComplete
        if self.scheduler.isActive():
            return
        self.scheduler.start(LexJob.ScheduleInitialDelay)  # Initiate chunking

    def stop(self):
        self.scheduler.stop()
        assert not self.scheduler.isActive()

    @benchmark
    def lexChunk(self, n: int = ChunkSize):
        assert not self.lexingComplete, "lexing complete, no need to keep chunking"
        assert self.lexGen is not None

        lexGen = self.lexGen
        tm = self.hqTokenMap
        ln = self.currentLine

        # Resume lexing current line
        tokens = tm[ln]

        try:
            for _i in range(n):
                ttype, text = next(lexGen)

                if '\n' not in text:
                    tokens.append((ttype, qstringLength(text)))
                else:
                    for part in text.split('\n'):
                        if part:
                            tm[ln].append((ttype, qstringLength(part)))
                        ln += 1
                        tm[ln] = []
                    ln -= 1  # for loop above inserts one too many lines
                    tokens = tm[ln]  # cache current line for next iteration

            self.currentLine = ln
            if self.requestedLine >= ln:
                assert not self.scheduler.isActive()
                self.scheduler.start(LexJob.ScheduleInterval)

        except StopIteration:
            self.currentLine = 0
            self.scheduler.stop()
            self.lqTokenMap = {}  # we won't need LQ tokens anymore - free some mem
            self.lexGen = None
            assert self.lexingComplete

        self.pulse.emit()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Built-in themes that (almost) don't depend on the desktop environment's style.

A theme is a bunch of color tokens that feed both a QPalette and theme.qss.
"""

from __future__ import annotations

import dataclasses
import enum
from contextlib import suppress
from pathlib import Path
from string import Template

from gitfourchette.qt import *
from gitfourchette.toolbox import mixColors, relativeLuminance


class ThemeName(enum.StrEnum):
    """
    Our built-in themes. These share the Prefs.qtStyle namespace with the
    native Qt style names (Breeze, Fusion, Windows...), so their values must
    not collide with anything QStyleFactory may return.
    """
    BuiltIn = "gitfourchette-builtin"


class ThemeAccent(enum.StrEnum):
    Blue = "#3daee9"
    Cyan = "#00d3b8"
    Green = "#3dd425"
    Yellow = "#e8cb2d"
    Orange = "#e9643a"
    Red = "#e93d58"
    Pink = "#e93a9a"
    Gray = "#686b6f"
    Indigo = "#926ee4"
    Purple = "#b875dc"


@dataclasses.dataclass
class ThemeColors:
    bg: str
    surface: str
    altRow: str
    border: str
    text: str
    hover: str
    selInactive: str
    button: str
    scrollHandle: str
    tooltipBg: str
    tooltipText: str
    danger: str = "red"

    accent: str = ThemeAccent.Blue
    outerRadius: int = 7
    innerRadius: int = round(outerRadius * .75)

    # All tokens below are inferred automatically. Do not define manually!
    onAccent: str = "white"
    defaultButton: str = "#f0f"
    defaultButtonHover: str = "#f0f"
    buttonHover: str = "#f0f"
    buttonPressed: str = "#f0f"
    tooltipBorder: str = "#f0f"
    textDim: str = "#f0f"
    textFaint: str = "#f0f"
    inputDisabled: str = "#f0f"
    light: str = "#f0f"
    midlight: str = "#f0f"
    mid: str = "#f0f"
    dark: str = "#f0f"
    shadow: str = "#f0f"

    # Engine-specific rounded rects
    menuRadius: int = outerRadius
    comboBoxMenuRadius: int = outerRadius

    # Engine-specific tokens. Meant to be prepended to a rule, e.g.:
    #     ${fusionOnly}QLabel {color: red}  /* red text only in Fusion */
    # The token is replaced with an empty string if the engine matches,
    # otherwise it's replaced with garbage so that the rule is ignored.
    fusionOnly: str = ""

    def __post_init__(self):
        """
        Derive intermediate colors automatically.
        """
        def mix(a: str, b: str, r=.5):
            return mixColors(QColor(a), QColor(b), r).name()

        # Determine whether 'onAccent' should be white or black.
        isDarkTheme = QColor(self.text).lightness() > QColor(self.surface).lightness()
        accentLuminance = relativeLuminance(QColor(self.accent))
        luminanceThreshold = .24 if isDarkTheme else .45

        # Menus, combobox lists, tooltips cannot be rounded with Fusion.
        # Breeze can round these, but draws menu shadows at a hardcoded radius,
        # which looks ugly if our radius is larger.
        engine = self.bestStyleEngine()
        maxMenuRadius = 7 if engine == "breeze" else 0

        self.defaultButton      = mix(self.button, self.accent, .25)
        self.defaultButtonHover = mix(self.button, self.accent, .33)
        self.buttonHover        = mix(self.button, self.text, .04)
        self.buttonPressed      = mix(self.button, self.accent, .66)
        self.tooltipBorder      = mix(self.tooltipText, self.tooltipBg, .8)
        self.textDim            = mix(self.text, self.surface, .4)
        self.textFaint          = mix(self.text, self.surface, .7)
        self.inputDisabled      = mix(self.bg, self.surface, .5)
        self.onAccent           = "white" if accentLuminance < luminanceThreshold else "black"

        self.fusionOnly         = "" if engine == "fusion" else "___IGNORE"
        self.menuRadius         = min(maxMenuRadius, self.outerRadius)
        self.comboBoxMenuRadius = min(maxMenuRadius, self.outerRadius)

        # Old-school 3D bevel/shadow colors. Derive from button.
        # Practically unused in Breeze, very rarely in Fusion, mostly in Windows.
        self.light              = QColor(self.button).lighter(200).name()
        self.midlight           = QColor(self.button).lighter(150).name()
        self.mid                = QColor(self.button).darker(150).name()
        self.dark               = QColor(self.button).darker(200).name()
        self.shadow             = QColor(self.button).darker(20).name()

    @classmethod
    def resolveTheme(cls, styleName: str, accent: QColor | None = None) -> ThemeColors | None:
        """
        Return the color tokens for one of our themes, or None if the input
        couldn't be parsed (e.g. if the given name is for a native Qt style).

        The input string must follow this format: "styleID,lightOrDark,#accentHex"
        Example: "gitfourchette-builtin,dark,#ff00ff"

        Omit lightOrDark and/or #accentHex to infer colors from system palette.
        """

        tokens = styleName.split(",")

        if tokens.pop(0) != ThemeName.BuiltIn:
            return None

        try:
            appScheme = QGuiApplication.styleHints().colorScheme()
            dark = appScheme == Qt.ColorScheme.Dark
        except AttributeError:  # Qt < 6.5
            dark = False

        while tokens:
            token = tokens.pop(0)
            if token in ("light", "dark"):
                dark = token == "dark"
            elif token.startswith("#"):
                accent = QColor(token)

        theme = MODERN_DARK if dark else MODERN_LIGHT
        if accent is not None:
            theme = dataclasses.replace(theme, accent=accent.name())

        return theme

    @classmethod
    def bestStyleEngine(cls) -> str:
        """
        Return the name of the Qt style engine upon which to base custom themes,
        as a lowercase string. Favor "breeze", if available, for better-looking
        menus (with drop shadows and rounded corners) that "fusion" cannot do.
        """
        hasBreeze = any(key.lower() == "breeze" for key in QStyleFactory.keys())  # noqa: SIM118
        return "breeze" if hasBreeze else "fusion"

    def buildStyleSheet(self) -> str:
        templatePath = Path(QFile("assets:style/theme.qss").fileName())
        templateText = templatePath.read_text(encoding="utf-8")
        replacements = dataclasses.asdict(self)
        qss = Template(templateText).substitute(replacements)
        return qss

    def buildPalette(self) -> QPalette:
        Role = QPalette.ColorRole

        normal = {
            Role.Window: self.bg,
            Role.WindowText: self.text,
            Role.Base: self.surface,
            Role.AlternateBase: self.altRow,
            Role.Text: self.text,
            Role.Button: self.button,
            Role.ButtonText: self.text,
            Role.BrightText: self.danger,
            Role.Highlight: self.accent,
            Role.HighlightedText: self.onAccent,
            Role.ToolTipBase: self.tooltipBg,
            Role.ToolTipText: self.tooltipText,
            Role.PlaceholderText: self.textFaint,
            Role.Link: self.accent,
            Role.LinkVisited: self.accent,
            Role.Light: self.light,
            Role.Midlight: self.midlight,
            Role.Mid: self.mid,
            Role.Dark: self.dark,
            Role.Shadow: self.shadow,
        }

        inactive = {
            Role.Highlight: self.selInactive,
            Role.HighlightedText: self.text,
        }

        disabled = {
            Role.WindowText: self.textFaint,
            Role.Text: self.textFaint,
            Role.ButtonText: self.textFaint,
            Role.Highlight: self.selInactive,
            Role.HighlightedText: self.textDim,
            Role.Base: self.inputDisabled,
            Role.Link: self.textDim,
        }

        with suppress(AttributeError):  # Qt 6.6+
            normal[Role.Accent] = self.accent

        palette = QPalette()

        for role, color in normal.items():
            palette.setColor(role, QColor(color))
        for role, color in inactive.items():
            palette.setColor(QPalette.ColorGroup.Inactive, role, QColor(color))
        for role, color in disabled.items():
            palette.setColor(QPalette.ColorGroup.Disabled, role, QColor(color))

        return palette


MODERN_DARK = ThemeColors(
    bg                = "#23262c",
    surface           = "#1b1e23",
    altRow            = "#1f2228",
    border            = "#444b55",
    text              = "#d7dbe1",
    hover             = "#12ffffff",
    selInactive       = "#343a44",
    button            = "#2c3138",
    scrollHandle      = "#2affffff",
    tooltipBg         = "#2f343c",
    tooltipText       = "#e4e7ec",
    danger            = "#ff6b60",
)

MODERN_LIGHT = ThemeColors(
    bg                = "#eef0f3",
    surface           = "#ffffff",
    altRow            = "#f7f8fa",
    border            = "#d2d6dd",
    text              = "#1f2329",
    hover             = "#10000000",
    selInactive       = "#dde1e8",
    button            = "#ffffff",
    scrollHandle      = "#38000000",
    tooltipBg         = "#2f343c",
    tooltipText       = "#f0f2f5",
    danger            = "#d92b1f",
)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import os
from collections.abc import Callable, Iterable, Iterator
from contextlib import suppress
from typing import Literal

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.exttools.usercommand import UserCommand
from gitfourchette.filelists.filebatchtask import FileBatchTask
from gitfourchette.filelists.filelistmodel import FileListModel
from gitfourchette.forms.searchbar import SearchBar
from gitfourchette.gitdriver import *
from gitfourchette.gitdriver.gitdeltafile import HexHashFFFF
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext, NavFlags
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.search.itemviewsearchprovider import ItemViewSearchProvider
from gitfourchette.settings import FileListClick, getDiffToolName, getExternalEditorName
from gitfourchette.tasks import *
from gitfourchette.toolbox import *


class FileListDelegate(QStyledItemDelegate):
    """
    Item delegate for QListView that supports highlighting search terms from a SearchBar
    """

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        widget = option.widget
        assert isinstance(widget, FileList)

        isActive = bool(option.state & QStyle.StateFlag.State_Active)
        isSelected = bool(option.state & QStyle.StateFlag.State_Selected)
        colorGroup = QPalette.ColorGroup.Active if isActive else QPalette.ColorGroup.Inactive

        # Gather data from model
        icon: QIcon = index.data(Qt.ItemDataRole.DecorationRole)
        emblem: QIcon | None = index.data(FileListModel.Role.Decoration2)
        font: QFont = index.data(Qt.ItemDataRole.FontRole)
        fullText: str = index.data(Qt.ItemDataRole.DisplayRole)
        searchTerm: str = widget.searchBar.provider.term()

        # Prepare icon and text rects
        rect = QRect(option.rect)
        rect.adjust(2, 0, -2, 0)

        # Begin painting
        painter.save()
        if font:
            painter.setFont(font)
        fontMetrics = painter.fontMetrics()

        # Draw default background
        widget.style().drawControl(QStyle.ControlElement.CE_ItemViewItem, option, painter, widget)

        # Draw icon
        for i in [icon, emblem]:
            if i is None:
                continue
            iconRect = QRect(rect)
            iconRect.setWidth(option.decorationSize.width())
            i.paint(painter, iconRect, option.decorationAlignment)
            rect.setLeft(iconRect.right() + 4)

        # Prepare elided text
        text = fontMetrics.elidedText(fullText, option.textElideMode, rect.width())

        # Split path into directory and filename for better readability
        isFileNameFirst = settings.prefs.pathDisplayStyle == PathDisplayStyle.FileNameFirst
        if isFileNameFirst:
            parts = text.split('\0', 1)
        else:
            slash = text.rfind('/')
            if slash < fullText.rfind('/'):
                # The last slash is either elided, or it's after the ellipsis.
                ellipsis = text.rfind('\u2026')
                slash = max(slash, ellipsis)
            parts = [text[:slash + 1], text[slash + 1:]]

        # Draw text parts
        textRectBackup = QRect(rect)
        for partNo, part in enumerate(parts):
            isDirectoryPart = isFileNameFirst == (partNo > 0)

            colorRole = (QPalette.ColorRole.HighlightedText if isSelected
                         else QPalette.ColorRole.PlaceholderText if isDirectoryPart
                         else QPalette.ColorRole.WindowText)

            textColor = option.palette.color(colorGroup, colorRole)
            if isSelected and isDirectoryPart:
                textColor.setAlphaF(0.7)

            painter.setPen(textColor)
            painter.drawText(rect, option.displayAlignment, part)

            # Prepare rect for next part
            partWidth = fontMetrics.horizontalAdvance(part)
            rect.setLeft(rect.left() + partWidth)

        # Restore text rect
        rect = textRectBackup

        # Highlight search term
        if searchTerm and searchTerm in fullText.lower():
            needlePos = text.lower().find(searchTerm)
            if needlePos < 0:
                needlePos = text.find("\u2026")  # unicode ellipsis character (...)
                needleLen = 1
            else:
                needleLen = len(searchTerm)

            SearchBar.highlightNeedle(painter, rect, text, needlePos, needleLen)

        # Finish painting
        painter.restore()


class FileList(QListView):
    nothingClicked = Signal()
    """ Only emitted if the widget has focus. """
    selectedCountChanged = Signal(int)
    openSubRepo = Signal(str)
    statusMessage = Signal(str)

    repoModel: RepoModel

    _selectionBackup: list[str]
    """
    Backup of selected paths before refreshing the view.
    """

    def __init__(self, repoModel: RepoModel, parent: QWidget, navContext: NavContext):
        super().__init__(parent)

        self.repoModel = repoModel

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onContextMenuRequested)

        flModel = FileListModel(self, self.repoModel.repo, navContext)
        self.setModel(flModel)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)

        self._selectionBackup = []

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        iconSize = self.fontMetrics().height()
        self.setIconSize(QSize(iconSize, iconSize))
        self.setTextElideMode(Qt.TextElideMode.ElideMiddle)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)  # prevent editing text after double-clicking
        self.setUniformItemSizes(True)  # potential perf boost with many files

        searchProvider = ItemViewSearchProvider(self)
        searchProvider.dataRole = FileListModel.Role.FilePath

        self.searchBar = SearchBar(self, searchProvider)
        self.searchBar.ui.forwardButton.hide()
        self.searchBar.ui.backwardButton.hide()
        self.searchBar.hide()
        flModel.modelAboutToBeReset.connect(self.searchBar.reevaluateSearchTerm)

        # Search result highlighter
        self.setItemDelegate(FileListDelegate(self))

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

        makeWidgetShortcut(self, self.searchBar.hideOrBeep, "Escape")
        makeWidgetShortcut(self, self.copyPaths, QKeySequence.StandardKey.Copy)

    def refreshPrefs(self):
        self.setVerticalScrollMode(settings.prefs.listViewScrollMode)
        nameFirst = settings.prefs.pathDisplayStyle == PathDisplayStyle.FileNameFirst
        self.setTextElideMode(Qt.TextElideMode.ElideRight if nameFirst else Qt.TextElideMode.ElideMiddle)

    @property
    def repo(self) -> Repo:
        return self.repoModel.repo

    @property
    def navContext(self) -> NavContext:
        return self.flModel.navContext

    @property
    def flModel(self) -> FileListModel:
        model = self.model()
        assert isinstance(model, FileListModel)
        return model

    def isEmpty(self):
        return self.model().rowCount() == 0

    def setContents(self, deltas: Iterable[GitDelta]):
        self.flModel.setContents(deltas)
        self.updateFocusPolicy()
        self.searchBar.reevaluateSearchTerm()

    def clear(self):
        self.flModel.clear()
        assert self.isEmpty()
        self.updateFocusPolicy()

    def updateFocusPolicy(self):
        focusPolicy = Qt.FocusPolicy.StrongFocus if not self.isEmpty() else Qt.FocusPolicy.ClickFocus
        self.setFocusPolicy(focusPolicy)

    # -------------------------------------------------------------------------
    # Context menu

    def makeContextMenu(self):
        deltas = list(self.selectedDeltas())
        if len(deltas) == 0:
            return None

        actions = self.contextMenuActions(deltas)
        menu = ActionDef.makeQMenu(self, actions)
        menu.setObjectName("FileListContextMenu")
        return menu

    def onContextMenuRequested(self, point: QPoint):
        menu = self.makeContextMenu()
        if menu is not None:
            menu.aboutToHide.connect(menu.deleteLater)
            menu.popup(self.mapToGlobal(point))

    def contextMenuActions(self, deltas: list[GitDelta]) -> list[ActionDef]:
        """ To be overridden """

        def pathDisplayStyleAction(pds: PathDisplayStyle):
            return ActionDef(
                englishTitleCase(trtables.enum(pds)),
                lambda: GFApplication.applyPrefs(pathDisplayStyle=pds),
                checkState=settings.prefs.pathDisplayStyle == pds)

        n = len(deltas)

        actions = [
            ActionDef.SEPARATOR,

            ActionDef(
                _n("Open &Folder", "Open {n} &Folders", n),
                self.showInFolder,
                "SP_DirIcon",
            ),

            ActionDef(
                _n("&Copy Path", "&Copy {n} Paths", n),
                self.copyPaths,
                shortcuts=QKeySequence.StandardKey.Copy,
            ),

            ActionDef(
                englishTitleCase(_("Path display style")),
                submenu=[pathDisplayStyleAction(style) for style in PathDisplayStyle],
            ),
        ]

        actions.extend(GFApplication.instance().mainWindow.contextualUserCommands(
            UserCommand.Token.File,
            UserCommand.Token.FileDir,
            UserCommand.Token.FileAbs,
            UserCommand.Token.FileDirAbs,
        ))

        return actions

    def contextMenuActionStash(self):
        return ActionDef(
            _("Stas&h Changes…"),
            self.wantPartialStash,
            icon="git-stash-black",
            shortcuts=TaskBook.shortcuts.get(NewStash, []))

    def contextMenuActionRevertMode(self, deltas: list[GitDelta], callback: Callable, ellipsis=True) -> ActionDef:
        n = len(deltas)
        action = ActionDef(_n("Revert Mode Change", "Revert Mode Changes", n), callback, enabled=False)

        # Scan deltas for mode changes
        for delta in deltas:
            # Scan for Modified, Renamed, or Copied
            if delta.status not in [GitStatus.Modified, GitStatus.Renamed, GitStatus.Copied]:
                continue

            # Skip if mode didn't change
            if delta.old.mode == delta.new.mode:
                continue

            # It has to be a mode we can actually revert
            if delta.new.mode not in [FileMode.BLOB, FileMode.BLOB_EXECUTABLE]:
                continue

            action.enabled = True

            # Set specific caption if it's a single item
            if n != 1:
                pass
            elif delta.new.mode == FileMode.BLOB_EXECUTABLE:
                action.caption = _("Revert Mode to Non-Executable")
            elif delta.new.mode == FileMode.BLOB:
                action.caption = _("Revert Mode to Executable")

        if ellipsis:
            action.caption += "…"

        return action

    def contextMenuActionsDiff(self, deltas: list[GitDelta]) -> list[ActionDef]:
        n = len(deltas)

        return [
            ActionDef(
                _("Open Diff in {0}", settings.getDiffToolName()),
                self.wantOpenInDiffTool,
                icon="vcs-diff"),

            ActionDef(
                _n("E&xport Diff As Patch…", "E&xport Diffs As Patch…", n),
                self.savePatchAs),
        ]

    def contextMenuActionsEdit(self, deltas: list[GitDelta]) -> list[ActionDef]:
        n = len(deltas)

        return [
            ActionDef(
                _("&Edit in {tool}", tool=settings.getExternalEditorName()),
                self.openWorkdirFile,
                icon="SP_FileIcon"),

            ActionDef(
                _n("Edit &HEAD Version in {tool}", "Edit &HEAD Versions in {tool}", n=n, tool=settings.getExternalEditorName()),
                self.openHeadRevision,
                enabled=any(not d.status.isAddedOrUntracked for d in deltas),
            ),
        ]

    def contextMenuActionBlame(self, deltas: list[GitDelta]) -> ActionDef:
        isEnabled = False
        if len(deltas) == 1:
            delta = deltas[0]
            assert self.navContext == NavContext.fromGitDeltaSource(delta.source)
            isEnabled = (not delta.source.isWorkdir()) or (not delta.status.isAddedOrUntracked)

        return ActionDef(
            englishTitleCase(OpenBlame.name()) + "\u2026",
            self.blameFile,
            icon=TaskBook.icons[OpenBlame],
            enabled=isEnabled,
            shortcuts=TaskBook.shortcuts[OpenBlame],
        )

    # -------------------------------------------------------------------------

    def confirmBatch(self, callback: FileBatchTask.UnitFunc, title: str, prompt: str):
        deltas = list(self.selectedDeltas())
        FileBatchTask.invoke(self, deltas, callback, title, prompt)

    def openWorkdirFile(self):
        def run(task: RepoTask, delta: GitDelta):
            entryPath = task.repo.in_workdir(delta.new.path)
            ToolProcess.startTextEditor(task.parentWidget(), entryPath)
            yield from task.flowEnterUiThread()  # dummy yield

        toolName = getExternalEditorName()

        self.confirmBatch(
            run,
            _("Open in {0}", toolName),
            _("Really open [# files] in {0}?", toolName))

    def wantOpenInDiffTool(self):
        def run(task: RepoTask, delta: GitDelta):
            yield from task.flowSubtask(OpenInDiffTool, delta)

        toolName = getDiffToolName()

        self.confirmBatch(
            run,
            _("Open in {0}", toolName),
            _("Really open [# files] in {0}?", toolName))

    def showInFolder(self):
        def run(task: RepoTask, delta: GitDelta):
            path = task.repo.in_workdir(delta.new.path)
            path = os.path.normpath(path)  # get rid of any trailing slashes (submodules)
            if not os.path.exists(path):  # check exists, not isfile, for submodules
                raise FileNotFoundError(_("File doesn’t exist at this path anymore."))
            showInFolder(path)
            yield from task.flowEnterUiThread()  # dummy yield

        self.confirmBatch(run, _("Open paths"), _("Really open [# folders]?"))

    def copyPaths(self):
        text = '\n'.join(self.repo.in_workdir(path) for path in self.selectedPaths())
        if not text:
            QApplication.beep()
            return

        if WINDOWS:  # Ensure backslash directory separators
            from pathlib import Path
            path = Path(text)
            text = str(path)

        QApplication.clipboard().setText(text)
        self.statusMessage.emit(clipboardStatusMessage(text))

    def selectRow(self, rowNumber=0):
        if self.model().rowCount() == 0:
            self.emitNothingClicked()
            self.clearSelection()
        else:
            self.setCurrentIndex(self.model().index(rowNumber or 0, 0))

    def emitNothingClicked(self):
        if self.hasFocus():
            self.nothingClicked.emit()

    def selectionChanged(self, justSelected: QItemSelection, justDeselected: QItemSelection):
        super().selectionChanged(justSelected, justDeselected)

        # We're the active FileList, clear counterpart.
        self._setCounterpart(-1)

        # Don't bother emitting signals if we're blocked
        if self.signalsBlocked():
            return

        selectedIndexes = self.selectedIndexes()
        numSelectedTotal = len(selectedIndexes)

        justSelectedIndexes = list(justSelected.indexes())
        if justSelectedIndexes:
            current = justSelectedIndexes[0]
        else:
            # Deselecting (e.g. with shift/ctrl) doesn't necessarily mean that the selection has been emptied.
            # Find an index that is still selected to keep the DiffView in sync with the selection.
            current = self.currentIndex()

            if current.isValid() and selectedIndexes:
                # currentIndex may be outside the selection, find the selected index that is closest to currentIndex.
                current = min(selectedIndexes, key=lambda index: abs(index.row() - current.row()))
            else:
                current = None

        self.selectedCountChanged.emit(numSelectedTotal)

        if not current or not current.isValid():
            self.emitNothingClicked()
            return

        locator: NavLocator = self.flModel.data(current, FileListModel.Role.Locator)
        locator = locator.withExtraFlags(NavFlags.BypassFileSelect)
        Jump.invoke(self, locator)

    def highlightCounterpart(self, loc: NavLocator):
        try:
            row = self.flModel.getRowForFile(loc.path)
        except KeyError:
            row = -1
        self._setCounterpart(row)

    def _setCounterpart(self, newRow: int):
        model = self.flModel
        oldRow = model.highlightedCounterpartRow

        if oldRow == newRow:
            return

        model.highlightedCounterpartRow = newRow

        if oldRow >= 0:
            oldIndex = model.index(oldRow, 0)
            self.update(oldIndex)

        if newRow >= 0:
            newIndex = model.index(newRow, 0)
            self.selectionModel().setCurrentIndex(newIndex, QItemSelectionModel.SelectionFlag.NoUpdate)
            self.update(newIndex)

    def mouseMoveEvent(self, event: QMouseEvent):
        """
        By default, ExtendedSelection lets the user select multiple items by
        holding down LMB and dragging. This event handler enforces single-item
        selection unless the user holds down Shift or Ctrl.
        """
        isLMB = bool(event.buttons() & Qt.MouseButton.LeftButton)
        isShift = bool(event.modifiers() & Qt.KeyboardModifier.ShiftModifier)
        isCtrl = bool(event.modifiers() & Qt.KeyboardModifier.ControlModifier)

        if isLMB and not isShift and not isCtrl:
            self.mousePressEvent(event)  # re-route event as if it were a click event
            self.scrollTo(self.indexAt(event.pos()))  # mousePressEvent won't scroll to the item on its own
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        super().mouseReleaseEvent(event)  # Let standard QListView selection occur first
        if event.button() == Qt.MouseButton.MiddleButton:
            self.onSpecialClick("middle")

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        super().mouseDoubleClickEvent(event)  # Let standard QListView selection occur first
        if event.button() == Qt.MouseButton.LeftButton:
            self.onSpecialClick("double")

    def onSpecialClick(self, click: Literal["middle", "double"]):
        if click == "double":
            action = settings.prefs.doubleClickFileList
        elif click == "middle":
            action = settings.prefs.middleClickFileList
        else:
            raise NotImplementedError(f"unknown special click kind '{click}'")

        if action == FileListClick.Nothing:
            pass
        elif action == FileListClick.Folder:
            self.showInFolder()
        elif action == FileListClick.Blame:
            self.blameFile()
        elif action == FileListClick.Edit:
            self.openWorkdirFile()
        elif action == FileListClick.DiffTool:
            self.wantOpenInDiffTool()
        elif action == FileListClick.Stage:
            self.wantStageOrUnstage()
        else:
            raise NotImplementedError(f"unknown special click action '{click}'")

    def selectedDeltas(self) -> Iterator[GitDelta]:
        for index in self.selectedIndexes():
            yield index.data(FileListModel.Role.Delta)

    def selectedPaths(self) -> Iterator[str]:
        for index in self.selectedIndexes():
            yield index.data(FileListModel.Role.FilePath)

    def earliestSelectedRow(self) -> int:
        try:
            i = iter(self.selectedIndexes())
            return next(i).row()
        except StopIteration:
            return -1

    def savePatchAs(self):
        deltas = list(self.selectedDeltas())
        ExportPatchCollection.invoke(self, deltas)

    def revertPaths(self):
        # TODO: Convert into a task? (So we can build the patch asynchronously)
        deltas = list(self.selectedDeltas())
        assert len(deltas) == 1
        delta = deltas[0]
        tokens = GitDriver.buildDiffCommand(delta)
        patchData = GitDriver.runSync(*tokens, directory=self.repo.workdir, strict=True)
        ApplyPatchData.invoke(self, patchData, reverse=True,
                              title=_("Revert changes in file"),
                              question=_("Do you want to revert this patch?"))

    def firstPath(self) -> str:
        index: QModelIndex = self.flModel.index(0)
        if index.isValid():
            return index.data(FileListModel.Role.FilePath)
        else:
            return ""

    def selectFile(self, file: str) -> bool:
        if not file:
            return False

        try:
            row = self.flModel.getRowForFile(file)
        except KeyError:
            return False

        if self.selectionModel().isRowSelected(row):
            # Re-selecting an already selected row may deselect it??
            return True

        self.selectRow(row)
        return True

    def deltaForFile(self, file: str) -> GitDelta:
        row = self.flModel.getRowForFile(file)
        return self.flModel.deltas[row]

    def openHeadRevision(self):
        def run(task: RepoTask, delta: GitDelta):
            fakeHeadFile = GitDeltaFile(delta.old.path, HexHashFFFF, source=GitDeltaSource.Commit, sourceCommit=self.repo.head_commit_id)
            fakeHeadDelta = GitDelta(GitStatus.Modified, new=fakeHeadFile)
            yield from task.flowSubtask(OpenRevisionInEditor, fakeHeadDelta, old=False)

        toolName = getExternalEditorName()
        self.confirmBatch(
            run,
            _("Open HEAD revision"),
            _("Really open [# files] in {0}?", toolName))

    def wantPartialStash(self):
        paths = set()
        for delta in self.selectedDeltas():
            # Add both old and new paths so that both are pre-selected
            # if we're stashing a rename.
            paths.add(delta.old.path)
            paths.add(delta.new.path)
        NewStash.invoke(self, list(paths))

    def wantStageOrUnstage(self):
        # To be overridden in DirtyFiles and StagedFiles
        QApplication.beep()

    def openSubmoduleTabs(self):
        for delta in self.selectedDeltas():
            if delta.isSubtreeCommitPatch():
                self.openSubRepo.emit(delta.new.path)

    def backUpSelection(self):
        oldSelected = list(self.selectedPaths())
        self._selectionBackup = oldSelected

    def clearSelectionBackup(self):
        self._selectionBackup = []

    def restoreSelectionBackup(self) -> bool:
        if not self._selectionBackup:
            return False

        paths = self._selectionBackup
        self._selectionBackup = []

        currentIndex: QModelIndex = self.currentIndex()
        cPath = currentIndex.data(FileListModel.Role.FilePath)

        if cPath not in paths:
            # Don't attempt to restore if we've jumped to another file
            return False

        if len(paths) == 1 and paths[0] == cPath:
            # Don't bother if the one file that we've selected is still the current one
            return False

        flModel = self.flModel
        selectionModel = self.selectionModel()
        SF = QItemSelectionModel.SelectionFlag

        with QSignalBlockerContext(self):
            # If we directly manipulate the QItemSelectionModel by calling .select() row-by-row,
            # then shift-selection may act counter-intuitively if the selection was discontiguous.
            # Preparing a QItemSelection upfront mitigates the strange shift-select behavior.
            newItemSelection = QItemSelection()
            for path in paths:
                with suppress(KeyError):
                    row = flModel.fileRows[path]
                    index = flModel.index(row, 0)
                    newItemSelection.select(index, index)
            selectionModel.clearSelection()
            selectionModel.select(newItemSelection, SF.Rows | SF.Select)
            selectionModel.setCurrentIndex(currentIndex, SF.Rows | SF.Current)

        return True

    def blameFile(self):
        def run(delta: GitDelta):
            if delta.source.isWorkdir():
                path = delta.old.path
                commit = NULL_OID
            else:
                path = delta.new.path
                commit = delta.new.sourceCommit
            OpenBlame.invoke(self, path, commit)

        # TODO: For now, only one blame at a time is supported
        try:
            delta = next(self.selectedDeltas())
            run(delta)
        except StopIteration:
            QApplication.beep()



# Form implementation generated from reading ui file 'remotedialog.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_RemoteDialog(object):
    def setupUi(self, RemoteDialog):
        RemoteDialog.setObjectName("RemoteDialog")
        RemoteDialog.setEnabled(True)
        RemoteDialog.resize(500, 178)
        RemoteDialog.setModal(True)
        self.formLayout = QFormLayout(RemoteDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.urlLabel = QLabel(parent=RemoteDialog)
        self.urlLabel.setObjectName("urlLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.urlLabel)
        self.urlLayout = QHBoxLayout()
        self.urlLayout.setObjectName("urlLayout")
        self.urlEdit = QLineEdit(parent=RemoteDialog)
        self.urlEdit.setObjectName("urlEdit")
        self.urlLayout.addWidget(self.urlEdit)
        self.protocolButton = ProtocolButton(parent=RemoteDialog)
        self.protocolButton.setText("https")
        self.protocolButton.setObjectName("protocolButton")
        self.urlLayout.addWidget(self.protocolButton)
        self.formLayout.setLayout(0, QFormLayout.ItemRole.FieldRole, self.urlLayout)
        self.nameLabel = QLabel(parent=RemoteDialog)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.nameEdit = QLineEdit(parent=RemoteDialog)
        self.nameEdit.setObjectName("nameEdit")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.nameEdit)
        self.skipFetchAllCheckBox = QCheckBox(parent=RemoteDialog)
        self.skipFetchAllCheckBox.setObjectName("skipFetchAllCheckBox")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.SpanningRole, self.skipFetchAllCheckBox)
        spacerItem = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.formLayout.setItem(3, QFormLayout.ItemRole.LabelRole, spacerItem)
        self.fetchAfterAddCheckBox = QCheckBox(parent=RemoteDialog)
        self.fetchAfterAddCheckBox.setChecked(True)
        self.fetchAfterAddCheckBox.setObjectName("fetchAfterAddCheckBox")
        self.formLayout.setWidget(4, QFormLayout.ItemRole.SpanningRole, self.fetchAfterAddCheckBox)
        self.buttonBox = QDialogButtonBox(parent=RemoteDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setCenterButtons(False)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(5, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.urlLabel.setBuddy(self.urlEdit)
        self.nameLabel.setBuddy(self.nameEdit)

        self.retranslateUi(RemoteDialog)
        self.buttonBox.accepted.connect(RemoteDialog.accept)
        self.buttonBox.rejected.connect(RemoteDialog.reject)
        RemoteDialog.setTabOrder(self.urlEdit, self.protocolButton)
        RemoteDialog.setTabOrder(self.protocolButton, self.nameEdit)
        RemoteDialog.setTabOrder(self.nameEdit, self.fetchAfterAddCheckBox)

    def retranslateUi(self, RemoteDialog):
        RemoteDialog.setWindowTitle(_p("RemoteDialog", "Edit remote"))
        self.urlLabel.setText(_p("RemoteDialog", "&URL:"))
        self.nameLabel.setText(_p("RemoteDialog", "&Name:"))
        self.skipFetchAllCheckBox.setToolTip(_p("RemoteDialog", "Tick this box to skip this remote during {cmd}. You can still fetch this remote explicitly, but it won’t be automatically fetched along with the others."))
        self.skipFetchAllCheckBox.setText(_p("RemoteDialog", "Skip when fetching &all remotes"))
        self.fetchAfterAddCheckBox.setText(_p("RemoteDialog", "&Fetch remote branches after adding"))
from gitfourchette.forms.protocolbutton import ProtocolButton



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import gc
import logging
import os
import sys
import textwrap
from pathlib import Path
from typing import Any

# Import as few internal modules as possible here to avoid premature initialization
# from cascading imports before the QApplication has booted.
from gitfourchette.localization import *
from gitfourchette.qt import *

if TYPE_CHECKING:
    from gitfourchette.forms.prefsdialog import PrefsDialog
    from gitfourchette.mainwindow import MainWindow
    from gitfourchette.settings import Session
    from gitfourchette.mount.mountmanager import MountManager
    from gitfourchette.sshagent import SshAgent

logger = logging.getLogger(__name__)


class GFApplication(QApplication):
    restyle = Signal()
    prefsChanged = Signal()
    regainForeground = Signal()
    fileDraggedToDockIcon = Signal(str)
    mouseSideButtonPressed = Signal(bool)

    # Lightweight state
    installedLocale: QLocale | None
    qtbaseTranslator: QTranslator
    tempDir: QTemporaryDir
    platformDefaultStyleName: str

    # Heavyweight state
    mainWindow: MainWindow | None
    initialSession: Session | None
    commandLinePaths: list
    sshAgent: SshAgent | None
    mountManager: MountManager | None

    @classmethod
    def instance(cls) -> GFApplication:
        me = QApplication.instance()
        assert isinstance(me, cls)
        return me

    @classmethod
    def main(cls):
        app = cls(sys.argv)
        app.beginSession()
        returnCode = app.exec()
        sys.exit(returnCode)

    def __init__(self, argv: list[str], barebones=False):
        super().__init__(argv)
        self.setObjectName("GFApplication")

        self.mainWindow = None
        self.initialSession = None
        self.commandLinePaths = []
        self.installedLocale = None
        self.qtbaseTranslator = QTranslator(self)
        self.sshAgent = None
        self.restylingGuard = 0

        # Show an error dialog in case of unhandled exceptions.
        # Note that debuggers may override the exception hook.
        self.injectExceptHook()

        # Don't use app.setOrganizationName because it changes QStandardPaths.
        self.setApplicationName(APP_SYSTEM_NAME)  # used by QStandardPaths
        self.setApplicationDisplayName(APP_DISPLAY_NAME)  # user-friendly name
        self.setApplicationVersion(APP_VERSION)
        self.setDesktopFileName(APP_IDENTIFIER)  # Wayland uses this to resolve window icons

        # Add assets search path
        if PYINSTALLER_MEIPASS:
            # PyInstaller, e.g. GitFourchette.app/Contents/Frameworks
            assetsParentPath = Path(PYINSTALLER_MEIPASS)
        else:
            # Relative to boot script
            assetsParentPath = Path(__file__ or sys.argv[0]).parent
        assetsSearchPath = str(assetsParentPath / "assets")
        QDir.addSearchPath("assets", assetsSearchPath)

        # Set app icon
        # - Except in macOS app bundles, which automatically use the embedded .icns file
        # - The file extension must be spelled out in some environments (e.g. Windows)
        if not (MACOS and APP_FREEZE_COMMIT):
            self.setWindowIcon(QIcon("assets:icons/gitfourchette.png"))

        # Get system default style & palette before applying further styling
        from gitfourchette.themes import ThemeName
        bootStyleName = self.style().objectName().lower()
        self.platformStandardAccent = self.palette().accent().color()
        if APP_TESTMODE and OFFSCREEN:
            # Don't force-set Qt style at the start of every offscreen test.
            # Don't touch default (Fusion) for pixel-perfect accuracy.
            assert bootStyleName == "fusion"
            self.platformDefaultStyleName = ""
        elif KDE and bootStyleName != "fusion":  # pragma: no cover
            # On KDE, be a good citizen and stick to system-provided theme
            # (unless we don't have anything better than Fusion, e.g. in the
            # AppImage's embedded Qt libraries)
            self.platformDefaultStyleName = bootStyleName
        else:  # pragma: no cover
            # On other platforms, default to a custom theme.
            self.platformDefaultStyleName = ThemeName.BuiltIn

        # Install translators for system language
        # (for command line parser to display localized text)
        self.installTranslators()

        self.createTempDir()

        self.installSigintHandler()

        # ---------------------------------------------------------------------
        # End of barebones initialization
        if barebones:
            return

        # Process command line
        parser = QCommandLineParser()
        parser.setApplicationDescription(qAppName() + " - " + _("The comfortable Git UI for Linux."))
        parser.addHelpOption()
        parser.addVersionOption()
        parser.addPositionalArgument("repos", _("Repository paths to open on launch."), "[repos...]")
        parser.process(argv)

        # Schedule cleanup on quit
        self.aboutToQuit.connect(self.endSession)

        from gitfourchette.globalshortcuts import GlobalShortcuts
        from gitfourchette.tasks import TaskBook

        # Prime singletons
        GlobalShortcuts.initialize()
        TaskBook.initialize()

        # Get initial session tabs
        commandLinePaths = parser.positionalArguments()
        commandLinePaths = [str(Path(p).resolve()) for p in commandLinePaths]
        self.commandLinePaths = commandLinePaths

    # -------------------------------------------------------------------------

    def createTempDir(self):
        path: str | Path = os.environ.get("GITFOURCHETTE_TEMPDIR", "")
        if path:
            pass
        elif FLATPAK:
            # Flatpak guarantees that "/run/user/1000/app/org.gitfourchette.gitfourchette" sits on a tmpfs,
            # and "/tmp" actually resolves to "/run/user/1000/app/org.gitfourchette.gitfourchette/tmp".
            # Use the real path instead of "/tmp" (QDir.tempPath() default value)
            # so that we can pass it to external tools outside our sandbox.
            path = Path(XDG_RUNTIME_DIR, "app", FLATPAK_ID)
            assert path.exists(), f"Expected to find Flatpak temp dir at: {path}"
        else:
            path = QDir.tempPath()

        path = Path(path, self.applicationName())
        tempDirTemplate = str(path)
        self.tempDir = QTemporaryDir(tempDirTemplate)
        self.tempDir.setAutoRemove(True)

    def injectExceptHook(self):
        # Show an error dialog in case of unhandled exceptions.
        # Note that debuggers may override the exception hook.
        sys.excepthook = self.customExceptHook

    @staticmethod
    def customExceptHook(exctype, value, tb):
        # Run default excepthook first
        sys.__excepthook__(exctype, value, tb)

        from gitfourchette.toolbox import excMessageBox
        excMessageBox(value, printExc=False)

    def installSigintHandler(self):
        import signal
        signal.signal(signal.SIGINT, self.onSigint)

        # Force Python interpreter to run every now and then so it can run the
        # Ctrl+C signal handler. Without this, the app won't actually die until
        # the window regains focus (https://stackoverflow.com/q/4938723).
        if __debug__:
            timer = QTimer(self)
            timer.start(300)
            timer.timeout.connect(lambda: None)

    def onSigint(self, *_dummy):
        # Deferring the quit to the next event loop gives the application some
        # time to wrap up the session.
        QTimer.singleShot(0, self.quit)
        # TODO: Actually return SIGINT from main

    # -------------------------------------------------------------------------

    def beginSession(self, bootUi=True):
        from gitfourchette.toolbox.messageboxes import NonCriticalOperation
        from gitfourchette import settings
        import pygit2

        # Make sure the temp dir exists
        tempDirPath = Path(self.tempDir.path())
        tempDirPath.mkdir(parents=True, exist_ok=True)

        # In a Flatpak, pygit2's XDG search path resolves to "~/.var/app/org.gitfourchette.gitfourchette/config/git"
        # by default. Users are much more likely to expect "~/.config/git" instead.
        if FLATPAK:
            userXdgGitDir = os.path.expanduser("~/.config/git")
            pygit2.settings.search_path[pygit2.enums.ConfigLevel.XDG] = userXdgGitDir

        # Load prefs file
        settings.prefs.reset()
        with NonCriticalOperation("Loading prefs"):
            settings.prefs.load()

        # Load history file
        settings.history.reset()
        with NonCriticalOperation("Loading history"):
            settings.history.load()
            settings.history.startups += 1
            settings.history.setDirty()

        # Set logging level from prefs
        self.dispatchSimplePrefsToStandaloneClasses()

        # Set language from prefs
        self.applyLanguagePref()

        # Load session file
        session = settings.Session()
        self.initialSession = session
        with NonCriticalOperation("Loading session"):
            session.load()
        if self.commandLinePaths:
            session.tabs += self.commandLinePaths
            session.activeTabIndex = len(session.tabs) - 1

        # Boot main window
        if bootUi:
            self.bootUi()

    def endSession(self, clearTempDir=True):
        self.removeEventFilter(self)  # Stop catching events

        from gitfourchette import settings
        from gitfourchette.syntax import LexJobCache
        from gitfourchette.toolbox.iconbank import clearStockIconCache
        if settings.prefs.isDirty():
            settings.prefs.write()
        if settings.history.isDirty():
            settings.history.write()
        self.stopSshAgent()

        assert not self.mountManager.mountedCommits, "mount points remaining"
        self.mountManager.deleteLater()
        self.mountManager = None

        LexJobCache.clear()  # don't cache lexed files across sessions (for unit testing)
        # RemoteLink.clearSessionPassphrases()  # don't cache passphrases across sessions (for unit testing)
        gc.collect()  # clean up Repository file handles (for Windows unit tests)
        if clearTempDir:
            clearStockIconCache()
            self.tempDir.remove()

    def bootUi(self):
        from gitfourchette.mainwindow import MainWindow
        from gitfourchette.toolbox import bquo
        from gitfourchette.settings import QtApiNames
        from gitfourchette.forms.donateprompt import DonatePrompt
        from gitfourchette.mount.mountmanager import MountManager

        assert self.mainWindow is None, "already have a MainWindow"
        assert self.initialSession is not None, "initial session should have been prepared before bootUi"

        # Initialize mountpoint manager
        self.mountManager = MountManager(self)

        self.applyQtStylePref()
        self.mainWindow = MainWindow()

        # Bind window signals
        self.regainForeground.connect(self.mountManager.checkAliveProcesses)
        self.regainForeground.connect(self.mainWindow.onRegainForeground)
        self.mainWindow.destroyed.connect(self.onMainWindowDestroyed)
        self.mouseSideButtonPressed.connect(self.mainWindow.onMouseSideButtonPressed)
        self.fileDraggedToDockIcon.connect(self.mainWindow.onFileDraggedToDockIcon)
        self.mountManager.mountPointsChanged.connect(self.mainWindow.fillGlobalMenuBar)
        self.mountManager.mountPointsChanged.connect(self.mainWindow.update)  # redraw GraphView
        self.mountManager.statusMessage.connect(self.mainWindow.statusBar2.showMessage)

        # To prevent flashing a window with incorrect dimensions,
        # restore the geometry BEFORE calling show()
        if not GNOME:  # Skip this on GNOME (issue #50)
            self.mainWindow.restoreGeometry(self.initialSession.windowGeometry)
        self.mainWindow.show()

        # Initialize SSH agent
        self.applySshAgentPref()

        # Restore session then consume it
        self.mainWindow.restoreSession(self.initialSession, self.commandLinePaths)
        self.initialSession = None
        self.commandLinePaths = []

        # Warn about incorrect Qt bindings
        if QT_BINDING_BOOTPREF and QT_BINDING_BOOTPREF.lower() != QT_BINDING.lower():  # pragma: no cover
            try:
                QtApiNames(QT_BINDING_BOOTPREF.lower())  # raises ValueError if not recognized
                text = _("Your preferred Qt binding {0} is not available on this machine. Using {1} instead.")
            except ValueError:
                text = _("Your preferred Qt binding {0} is not recognized by {app}. Using {1} instead. (Supported values: {known})")
            text = text.format(bquo(QT_BINDING_BOOTPREF), bquo(QT_BINDING.lower()), app=qAppName(),
                               known=", ".join(e for e in QtApiNames if e))

            QMessageBox.information(self.mainWindow, _("Qt binding unavailable"), text)

        DonatePrompt.onBoot(self.mainWindow)

        self.installEventFilter(self)

    def onMainWindowDestroyed(self):
        logger.debug("Main window destroyed")
        self.mainWindow = None
        self.removeEventFilter(self)  # Stop catching events

    # -------------------------------------------------------------------------

    def installTranslators(self, preferredLanguage: str = ""):
        if preferredLanguage:
            locale = QLocale(preferredLanguage)
        elif APP_TESTMODE:
            # Fall back to English in unit tests regardless of the host machine's locale
            # because many unit tests look for pieces of text in dialogs.
            locale = QLocale(QLocale.Language.English)
        else:  # pragma: no cover
            locale = QLocale()  # "Automatic" setting: Get system locale

        # Force English locale for RTL languages. RTL support isn't great for now,
        # and we have no localizations for RTL languages yet anyway.
        if locale.textDirection() != Qt.LayoutDirection.LeftToRight:
            locale = QLocale(QLocale.Language.English)

        # Set default locale
        QLocale.setDefault(locale)
        previousLocale = self.installedLocale
        self.installedLocale = locale

        if previousLocale is not None and locale.name() == previousLocale.name():
            # Previous locale is similar enough to new locale. Don't reload translators.
            return

        # Try to load gettext translator for application strings.
        self._installGettextTranslator(locale)

        # Remove previously installed qtbase translator
        QCoreApplication.removeTranslator(self.qtbaseTranslator)

        # Load qtbase translator
        if not QT5:  # Qt 5 doesn't have QLibraryInfo.path
            qtTranslationsDir = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
            if self.qtbaseTranslator.load(locale, "qtbase", "_", qtTranslationsDir, ".qm"):
                QCoreApplication.installTranslator(self.qtbaseTranslator)

    @staticmethod
    def _installGettextTranslator(locale: QLocale):
        # We'll be resolving the path to an '.mo' file that best matches the given locale.
        moFilePath = ""

        # Match Qt language code with .mo files exported from Weblate.
        # The codes for most languages already match, but Chinese is a notable exception.
        qtLocaleToWeblate = {
            "zh_CN": "zh_Hans",
        }

        languageCode = locale.name()
        languageCode = qtLocaleToWeblate.get(languageCode, languageCode)

        # Look for a territory-specific file first, then fall back to a
        # generic language file (e.g. 'fr_CA' then 'fr').
        try:
            genericLanguageCode = QLocale.languageToCode(locale.language())
        except AttributeError:  # pragma: no cover - Compatibility with Qt 5 and pre-Qt 6.3
            genericLanguageCode = languageCode.split("_")[0]

        for stem in languageCode, genericLanguageCode:
            languageFile = QFile(f"assets:lang/{stem}.mo")
            if languageFile.exists():
                moFilePath = languageFile.fileName()
                break

        # Install translations from the '.mo' file.
        # If we couldn't find a file, this will fall back to American English.
        installGettextTranslator(moFilePath)

    # -------------------------------------------------------------------------

    @classmethod
    def applyPrefs(cls, **kwargs):
        cls.instance()._applyPrefs(kwargs)

    def _applyPrefs(self, prefDiff: dict[str, Any], writeNow=False):
        from gitfourchette.settings import prefs

        # Reset "don't show again" dialogs
        if prefDiff.get("resetDontShowAgain", False):
            prefDiff["resetDontShowAgain"] = False
            prefDiff["dontShowAgain"] = []

        # Invalidate stale per-repo ref sorting settings
        if "refSort" in prefDiff:
            prefDiff["refSortClearTimestamp"] = QDateTime.currentSecsSinceEpoch()

        # Commit changes from prefDiff to the actual prefs
        dirty = False
        realPrefDict = prefs.__dict__
        for key, newValue in prefDiff.items():
            try:
                oldValue = realPrefDict[key]
            except KeyError as ex:
                raise KeyError(f"invalid pref key {key}") from ex

            if oldValue == newValue:  # no-op
                pass

            oldType = type(oldValue)
            newType = type(newValue)
            if oldType != newType:
                raise TypeError(f"{key} type mismatch: expected {oldType}, got {newType}")

            realPrefDict[key] = newValue
            dirty = True

        # Early out if the prefs didn't change
        if not dirty:
            return

        prefs.setDirty()
        if writeNow:
            prefs.write()

        # ---------------------------------------------------------------------
        # Heed new settings in GFApplication

        assert self.mainWindow is not None

        self.dispatchSimplePrefsToStandaloneClasses()

        if "qtStyle" in prefDiff:
            self.applyQtStylePref()

        if "language" in prefDiff:
            self.applyLanguagePref()

        resetSshAgent = "ownSshAgent" in prefDiff
        # Flatpak: If git's sandboxed state changes, we need to recreate
        # ssh-agent to be (non-)sandboxed accordingly
        resetSshAgent |= FLATPAK and "gitPath" in prefDiff
        if resetSshAgent:
            self.applySshAgentPref()

        # ---------------------------------------------------------------------
        # Notify widgets

        changedKeys = set(prefDiff.keys())
        self.prefsChanged.emit()
        self.mainWindow.onApplyPrefs(changedKeys)

    def dispatchSimplePrefsToStandaloneClasses(self):
        from gitfourchette import settings
        from gitfourchette.gitdriver import GitDriver
        from gitfourchette.toolbox.fittedtext import FittedText

        logging.root.setLevel(settings.prefs.verbosity.value)
        GitDriver.setGitPath(settings.prefs.gitPath)
        FittedText.enable = settings.prefs.condensedFonts

    def applyLanguagePref(self):
        from gitfourchette import settings
        from gitfourchette import trtables
        from gitfourchette.tasks.taskbook import TaskBook

        self.installTranslators(settings.prefs.language)

        # Regenerate rosetta stones
        trtables.retranslate(settings.prefs.language)
        TaskBook.retranslate()

    def applyQtStylePref(self, paletteOnly=False):
        self.restylingGuard += 1
        try:
            self._applyQtStylePref(paletteOnly)
        finally:
            self.restylingGuard -= 1

    def _applyQtStylePref(self, paletteOnly=False):
        from gitfourchette import settings
        from gitfourchette.syntax.colorscheme import ColorScheme
        from gitfourchette.toolbox import mixColors, iconbank
        from gitfourchette.themes import ThemeColors

        effectiveStyle = settings.prefs.qtStyle

        if not effectiveStyle:
            effectiveStyle = self.platformDefaultStyleName

        # See if it's a custom theme
        accent = self.platformStandardAccent
        customTheme = ThemeColors.resolveTheme(effectiveStyle, accent)
        if customTheme:
            effectiveStyle = ThemeColors.bestStyleEngine()
            palette = customTheme.buildPalette()
        else:
            palette = QPalette()
        # Set custom palette, or reset standard palette
        self.setPalette(palette)

        # ----------------------------------------------------------------------
        # Build stylesheet

        qss = Path(QFile("assets:style/base.qss").fileName()).read_text()

        # Palette-dependent styling
        windowColor = self.palette().color(QPalette.ColorRole.Window)
        textColor = self.palette().color(QPalette.ColorRole.Text)
        headerBg = mixColors(windowColor, textColor, .07)
        headerFg = mixColors(windowColor, textColor, .82)
        faintSep = mixColors(windowColor, textColor, .18)
        qss += textwrap.dedent(f"""
            ContextHeader {{ background-color: {headerBg.name()}; }}
            ContextHeader QLabel {{ color: {headerFg.name()}; }}
            QFaintSeparator {{ background: {faintSep.name()}; color: transparent; }}
        """)

        if MACOS:  # Strip iOS-y QMessageBox styling (all bold)
            qss += "QMessageBox QLabel {font-weight: normal;}"

        # Append our own theme, if any (its rules take precedence over the above)
        if customTheme is not None:
            qss += customTheme.buildStyleSheet()

        # Install the stylesheet. It's OK to re-evaluate the same stylesheet if
        # the QSS is identical (needed to react to system palette changes)
        self.setStyleSheet(qss)

        # ----------------------------------------------------------------------

        # Set Qt style
        if effectiveStyle and not paletteOnly:
            self.setStyle(effectiveStyle)

        if MACOS:
            self.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus, settings.qtIsNativeMacosStyle())

        # ----------------------------------------------------------------------

        # Force RecolorSvgIconEngine to re-render the icons
        iconbank.clearStockIconCache()
        QPixmapCache.clear()

        # Reset syntax highlighting fallback
        ColorScheme.refreshFallbackScheme()

    def applySshAgentPref(self):
        from gitfourchette import settings
        from gitfourchette.sshagent import SshAgent
        from gitfourchette.toolbox import showWarning, paragraphs

        wantAgent = settings.prefs.ownSshAgent
        sandbox = settings.prefs.isGitSandboxed()

        if not wantAgent:
            self.stopSshAgent()
            return

        if self.sshAgent and sandbox != self.sshAgent.isSandboxed():
            # Sandboxed state of the current agent isn't what we want, restart it
            self.stopSshAgent()

        if self.sshAgent:
            # Already have an agent
            return

        try:
            self.sshAgent = SshAgent(self, sandbox=sandbox)
        except Exception as exc:
            message = paragraphs(
                _("Couldn’t start {0} ({1}).", "ssh-agent", str(exc)),
                _("Please make sure OpenSSH is installed on your system."))
            showWarning(self.mainWindow, _("SSH agent"), message)

    def stopSshAgent(self):
        if self.sshAgent:
            self.sshAgent.stopAndWait()
            self.sshAgent.deleteLater()
            self.sshAgent = None

    # -------------------------------------------------------------------------

    def processEventsNoInput(self):
        self.processEvents(QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)

    def eventFilter(self, watched, event: QEvent):
        assert self.mainWindow, "eventFilter shouldn't catch events post-mortem"

        eventType = event.type()
        isPress = eventType == QEvent.Type.MouseButtonPress

        if eventType == QEvent.Type.FileOpen:
            # Called if dragging something to dock icon on macOS.
            # Ignore in test mode - the test runner may send a bogus FileOpen before we're ready to process it.
            assert isinstance(event, QFileOpenEvent)
            path = event.file()
            if not APP_TESTMODE:
                self.fileDraggedToDockIcon.emit(path)

        elif eventType == QEvent.Type.ApplicationStateChange:
            # Refresh current RepoWidget when the app regains the active state (foreground)
            if QGuiApplication.applicationState() == Qt.ApplicationState.ApplicationActive:
                QTimer.singleShot(0, self.regainForeground)

        elif isPress or eventType == QEvent.Type.MouseButtonDblClick:
            # As of PyQt6 6.8, QContextMenuEvent sometimes pretends that its event type is a MouseButtonDblClick
            if PYQT6 and not isinstance(event, QMouseEvent):
                logger.warning(f"QContextMenuEvent pretends it's a double click: {event}")
                return False

            # Intercept back/forward mouse clicks
            assert isinstance(event, QMouseEvent)
            button = event.button()
            isBack = button == Qt.MouseButton.BackButton
            isForward = button == Qt.MouseButton.ForwardButton
            if isBack or isForward:
                if isPress:
                    self.mouseSideButtonPressed.emit(isForward)
                # Eat clicks or double-clicks of back and forward mouse buttons
                return True

        elif eventType == QEvent.Type.PaletteChange and watched is self.mainWindow:
            # Recolor some widgets when palette changes (light to dark or vice-versa).
            if not self.restylingGuard:
                # Save new accent color from environment before tweaking palette (e.g. KDE color change)
                self.platformStandardAccent = self.palette().accent().color()
                try:
                    self.applyQtStylePref(paletteOnly=True)
                finally:
                    self.restyle.emit()  # tell other widgets about it

        elif eventType == QEvent.Type.StatusTip:
            # Eat QStatusTipEvent. The menubar emits those when a menu is hovered;
            # but since we don't use status tips, the status bar is cleared for no reason.
            if APP_DEBUG:
                tip: str = event.tip() # type: ignore[attr-defined] # incomplete stubs
                assert not tip, "assuming QStatusTipEvent is always empty"
            return True

        elif eventType == QEvent.Type.Show and isinstance(watched, QDialog):
            self.installDialogReturnShortcut(watched)

        return False

    # -------------------------------------------------------------------------
    # Utilities

    def openPrefsDialog(self, focusOnPrefKey: str = "") -> PrefsDialog:
        from gitfourchette.forms.prefsdialog import PrefsDialog

        dlg = PrefsDialog(self.mainWindow, focusOnPrefKey)
        dlg.accepted.connect(lambda: self._applyPrefs(dlg.prefDiff, writeNow=True))
        dlg.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)  # don't leak dialog
        dlg.show()
        return dlg

    @staticmethod
    def installDialogReturnShortcut(dialog: QDialog):
        """
        In KDE, the Return key typically triggers a QDialog's default button
        ("OK") regardless of the widget that has keyboard focus.

        However, in other desktop environments like GNOME or Cinnamon, Return
        triggers the QRadioButton or QCheckBox that has keyboard focus (in
        addition to Space), preventing the QDialog from being accepted.

        This function overrides DE-specific behavior and makes the Return key
        accept the dialog consistently, like in KDE.

        Call this *after* the dialog has been shown to avoid conflicts with any
        shortcuts that the desktop environment may want to install.
        """

        assert dialog.isVisible(), "call QDialog.show() first"

        # Skip dialogs we've already seen.
        attr = "_guard_installDialogReturnShortcut"
        if getattr(dialog, attr, False):
            return
        setattr(dialog, attr, True)

        # Don't tamper with shortcuts in environments where the native behavior
        # is adequate.
        if (KDE or MACOS) and not OFFSCREEN:
            return

        buttonBox = dialog.findChild(QDialogButtonBox)
        if not buttonBox:
            return

        okButton = buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        if not okButton:
            return

        # Don't conflict with any shortcuts that may have been installed
        # by the desktop environment after QDialog.show().
        # For example, KDE installs its own Ctrl+Return shortcut.
        if okButton.findChild(QShortcut):
            return

        from gitfourchette.toolbox.qtutils import makeWidgetShortcut

        # "Return" is the main key, "Enter" is the numpad key.
        makeWidgetShortcut(
            okButton, okButton.click,
            "Ctrl+Return", "Ctrl+Enter",
            "Return", "Enter",
            context=Qt.ShortcutContext.WindowShortcut)

        logger.debug(f"Installed return shortcut on {dialog}")



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.signatureform import SignatureOverride
from gitfourchette.forms.ui_commitdialog import Ui_CommitDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class CommitDialog(QDialog):
    InfoIconSize = 16

    @property
    def acceptButton(self) -> QPushButton:
        return self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

    def __init__(
            self,
            initialText: str,
            authorSignature: Signature,
            committerSignature: Signature,
            amendingCommitHash: str,
            detachedHead: bool,
            repositoryState: RepositoryState,
            emptyCommit: bool,
            gpgFlag: bool,
            gpgKey: str,
            hooks: list[str],
            parent: QWidget):
        super().__init__(parent)

        self.originalAuthorSignature = authorSignature
        self.originalCommitterSignature = committerSignature

        self.ui = Ui_CommitDialog()
        self.ui.setupUi(self)
        self.ui.gpg.setup(gpgFlag, gpgKey)
        self.ui.hookButton.setup(hooks)

        for button in (self.ui.gpg, self.ui.signoffButton, self.ui.hookButton):
            button.setAutoRaise(True)

        self.ui.signatureButton.setIcon(stockIcon("view-visible"))

        self.ui.signature.setSignature(authorSignature)
        self.ui.signature.signatureChanged.connect(self.refreshSignaturePreview)

        # Make summary text edit font larger
        tweakWidgetFont(self.ui.summaryEditor, 130)

        if amendingCommitHash:
            prompt = _("Amend commit message")
            buttonCaption = _("A&mend")
            self.setWindowTitle(_("Amend Commit {0}", amendingCommitHash))
        else:
            prompt = _("Enter commit summary")
            buttonCaption = _("Co&mmit")
            self.setWindowTitle(englishTitleCase(_("New commit")))

        warning = ""
        if repositoryState == RepositoryState.MERGE:
            warning = _("This commit will conclude the merge.")
        elif repositoryState == RepositoryState.CHERRYPICK:
            warning = _("This commit will conclude the cherry-pick.")
        elif repositoryState == RepositoryState.REVERT:
            warning = _("This commit will conclude the revert.")
        elif amendingCommitHash and emptyCommit:
            warning = _("Amending commit {0} without changing its contents (no staged changes).",
                        lquo(amendingCommitHash))
        elif amendingCommitHash:
            warning = _("Amending commit {0}.", lquo(amendingCommitHash))
        elif detachedHead:
            warning = _("<b>Detached HEAD</b> – You are not in any branch! "
                        "You should create a branch to keep track of your commit.")
        elif emptyCommit:
            warning = _("Creating an empty commit (no staged changes).")

        self.ui.infoBox.setVisible(bool(warning))
        self.ui.infoText.setText(warning)
        self.ui.infoIcon.setPixmap(stockIcon("SP_MessageBoxInformation").pixmap(self.InfoIconSize))
        self.ui.infoIcon.setMaximumWidth(self.InfoIconSize)

        self.acceptButton.setText(buttonCaption)
        self.ui.summaryEditor.setPlaceholderText(prompt)

        self.ui.counterLabel.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.ui.counterLabel.setMinimumWidth(self.ui.counterLabel.fontMetrics().horizontalAdvance('000'))
        self.ui.counterLabel.setMaximumWidth(self.ui.counterLabel.minimumWidth())

        self.validator = ValidatorMultiplexer(self)
        self.validator.setGatedWidgets(self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok))
        self.validator.connectInput(self.ui.summaryEditor, self.hasNonBlankSummary, showError=False)
        self.ui.signature.installValidator(self.validator)

        self.ui.summaryEditor.textEdited.connect(self.sanitizeLineBreaksInSummary)
        self.ui.summaryEditor.textChanged.connect(self.updateCounterLabel)
        self.ui.revealSignature.checkStateChanged.connect(lambda: self.validator.run())
        self.ui.revealSignature.checkStateChanged.connect(lambda: self.refreshSignaturePreview())

        split = initialText.split('\n', 1)
        if len(split) >= 1:
            self.ui.summaryEditor.setText(split[0].strip())
        if len(split) >= 2:
            self.ui.descriptionEditor.setPlainText(split[1].strip())

        self.ui.revealSignature.setChecked(False)
        self.ui.signatureBox.setVisible(False)

        self.updateCounterLabel()
        self.validator.run()
        self.refreshSignaturePreview()

        # Focus on summary editor before showing
        self.ui.summaryEditor.setFocus()

        # Pack the layout
        packDialog(self)
        self.resize(self.width(), self.minimumSizeHint().height())

    def sanitizeLineBreaksInSummary(self, text: str):
        if '\n' not in text:
            return

        # Walk back undo stack to get rid of the operation that inserted linebreaks.
        # This way, this callback won't run again when the user presses Ctrl+Z.
        with QSignalBlockerContext(self.ui.summaryEditor):
            self.ui.summaryEditor.undo()

        summary, details = text.split('\n', 1)
        summary = summary.rstrip()
        details = details.lstrip()

        # This pushes the new text to the QLineEdit's undo stack
        # (whereas setText clears the undo stack).
        self.ui.summaryEditor.selectAll()
        self.ui.summaryEditor.insert(summary)
        self.ui.descriptionEditor.selectAll()
        self.ui.descriptionEditor.insertPlainText(details)

    def updateCounterLabel(self):
        text = self.ui.summaryEditor.text()
        self.ui.counterLabel.setText(str(len(text)))

    def hasNonBlankSummary(self, text):
        if bool(text.strip()):
            return ""
        else:
            return _("Cannot be empty.")

    def getFullMessage(self) -> str:
        summary = self.ui.summaryEditor.text()
        details = self.ui.descriptionEditor.toPlainText()

        hasDetails = bool(details.strip())

        if not hasDetails:
            message = summary
        else:
            message = f"{summary}\n\n{details}"

        if not message:
            return ""

        # Vanilla git enforces a final newline in commit messages whether we want it or not,
        # so just add one ourselves so that we get the same outcome with libgit2.
        if not message.endswith("\n"):
            message += "\n"

        return message

    def getOverriddenSignatureKind(self):
        if self.ui.revealSignature.isChecked() and self.ui.signature.getSignature():
            return self.ui.signature.replaceWhat()
        return SignatureOverride.Nothing

    def getOverriddenAuthorSignature(self):
        if self.getOverriddenSignatureKind() in [SignatureOverride.Author, SignatureOverride.Both]:
            return self.ui.signature.getSignature()
        else:
            return None

    def getOverriddenCommitterSignature(self):
        if self.getOverriddenSignatureKind() in [SignatureOverride.Committer, SignatureOverride.Both]:
            return self.ui.signature.getSignature()
        else:
            return None

    def refreshSignaturePreview(self):
        def formatSignatureForToolTip(sig: Signature):
            if sig is None:
                return "???"
            dateText = signatureDateFormat(sig)
            return f"{escape(sig.name)} &lt;{escape(sig.email)}&gt;<br><small>{escape(dateText)}</small>"

        author = self.getOverriddenAuthorSignature() or self.originalAuthorSignature
        committer = self.getOverriddenCommitterSignature() or self.originalCommitterSignature

        muted = mutedToolTipColorHex()

        tt = "<p style='white-space: pre'>"
        tt += f"<span style='color: {muted}'>" + _("Authored by:") + "</span> "
        tt += formatSignatureForToolTip(author)
        if not signatures_equalish(author, self.originalAuthorSignature):
            tt += "\n<span style='font-weight: bold'>" + _("(overridden manually)") + "</span>"

        tt += f"\n\n<span style='color: {muted}'>" + _("Committed by:") + "</span> "
        tt += formatSignatureForToolTip(committer)
        if not signatures_equalish(committer, self.originalCommitterSignature):
            tt += "\n<span style='font-weight: bold'>" + _("(overridden manually)") + "</span>"

        self.ui.signatureButton.setToolTip(tt)
        self.ui.revealSignature.setToolTip(tt)
